/*
 * SCTInterface.java
 *
 * Created on 15 November 2002, 12:01
 */

package GuiComponents.System;

import ipc.*;
import java.util.*;
import java.util.logging.*;
import SctApiTest.*;
import Sct_CalibrationController.*;
import Sct_SctApi.*;
import AnalysisServiceI.*;
import sctConf.*;
import FitterI.*;


/**
 * This class is an interface to the rest of the Software System
 * It contains methods to get access to all the various components that
 * may or may not be running.
 *
 * Users should probably not cache references for long periods of time
 *
 * This class and ISInterface mirror some of the methods from SctNames in the C++ code
 *
 * @author Matthew Palmer
 */
public class SystemInterface {    
    
    /**
     * Get the singleton interface.
     */
    public static SystemInterface getInstance() {
        return instance;
    }

    /**
     * Re-obtain all connections to external processes.
     */
    public void refresh() {
        p = new Partition(partitionName);                
        testapi = null;
        cc = null; 
        api = null;
        config = null;
        analysis = null;
        fitter = null;
    }

    /**
     * Get the IPC Partition that all our processes are contained in.
     */
    public Partition getPartition() {
        return p;
    }
    
    /**
     * Get a reference to the SCTTestAPI
     */
    public sctapi getTestApi() {
        if (testapi == null) {
            try {
                org.omg.CORBA.Object ob = p.lookup("SCTTestAPIServerImpl", "SCTTestAPIServer");                
                testapi = sctapiHelper.narrow(ob);
		if (testapi == null) {
		    logger.info("Failed to lookup SCTTestAPI using: SCTTestAPIServerImpl and SCTTestAPIServer.");   
		}
            } catch (org.omg.CORBA.SystemException se) {                		
		logger.info("Failed to lookup SCTTestAPI using: SCTTestAPIServerImpl and SCTTestAPIServer. Exception: " + se.getMessage());
            }
        }
        return testapi;
    }
    
    /**
     * Gets a reference to the CalibrationController
     */
    public CalibrationController getCalibrationController() {
        if (cc == null) {
            try {
                org.omg.CORBA.Object ob = p.lookup(CalibrationController.instanceName, CalibrationController.serverName); 
                cc = CalibrationControllerHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup CalibrationController using: " + CalibrationController.instanceName + " and " + CalibrationController.serverName + ". Exception: " + se.getMessage());
            }
        }
        return cc;        
    }
    
    /**
     * Gets a reference to the Api
     */
    public SctApi getSctApi() {
        if (api == null) {
            try {
                org.omg.CORBA.Object ob = p.lookup("SCTSctApiServer","Test_Server"); 
                api = SctApiHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup AnalysisService using: SCTSctApiServer and Test_Server . Exception: " + se.getMessage());
            }
        }
        return api;        
    }
    
    /**
     * Gets a reference to the configuration service
     */
    public Configuration getConfigurationService() {
        if (config == null) {
            try {
                org.omg.CORBA.Object ob = p.lookup("SCTConfigurationServer", "Test_Server"); 
                config = ConfigurationHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup Configuration using: SCTConfigurationServer and Test_Server. Exception: " + se.getMessage());
            }
        }
        return config;        
    }
    
    /**
     * Gets a reference to the analysis service
     */
    public AnalysisServiceInterface getAnalysisService() {
        if (analysis == null) {
            try {
                org.omg.CORBA.Object ob = p.lookup(AnalysisServiceInterface.instanceName, AnalysisServiceInterface.serverName); 
                analysis = AnalysisServiceInterfaceHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup AnalysisService using: " + AnalysisServiceInterface.instanceName + " and " + AnalysisServiceInterface.serverName + ". Exception: " + se.getMessage());
            }
        }
        return analysis;                
    }
    
    /**
     * Gets a reference to the fitting service
     */
    public FitterInterface getFittingService() {
        if (fitter == null) {
            try {
                org.omg.CORBA.Object ob = p.lookup(FitterInterface.instanceName, FitterInterface.serverName); 
                fitter = FitterInterfaceHelper.narrow(ob);
            } catch (org.omg.CORBA.SystemException se) {                
                logger.info("Failed to lookup Fitter using: " + FitterInterface.instanceName + " and " + FitterInterface.serverName + ". Exception: " + se.getMessage());
            }
        }
        return fitter;                
    }
    
    /** Creates a new instance of SCTInterface */
    private SystemInterface() {
        logger = Logger.getLogger("GuiComponents.System");     
        refresh();
    }            
    
    private static SystemInterface instance = new SystemInterface();
    private static final String partitionName = "SCT";
    private Partition p;
    private Logger logger;    
    private sctapi testapi;
    private CalibrationController cc; 
    private SctApi api;
    private Configuration config;
    private AnalysisServiceInterface analysis;
    private FitterInterface fitter;
}
