#include "RawScanResultStreamer.h"
#include "Sct/SctNames.h"
#include "Sct/SctParameters.h"
#include "Sct/LogicErrors.h"
#include "Sct/VersionNotSupportedException.h"
#include "../RawScanResult.h"
#include "SctApi/dataTypes.h"

#include "TH2.h"

#include <iostream>
#include <sstream>

using namespace std;
using namespace Sct;

namespace SctData {
namespace IO {

RawScanResultStreamer::RawScanResultStreamer() throw() {}

bool RawScanResultStreamer::inMap = IOManager::addToMap("SctData::RawScanResult", auto_ptr<Streamer>(new RawScanResultStreamer()));

shared_ptr<Streamable> RawScanResultStreamer::read(IStream& in, const IOManager& manager) const throw(LogicError, IoError) {
    shared_ptr<Streamable> ad (&helper.create());
    read(in, *ad, manager);
    return ad;
}

void RawScanResultStreamer::write(OStream& out, const Streamable& ob, const IOManager& manager) const throw(LogicError, IoError) {
    //Let superclass do its worker
    ScanResultStreamer::write(out, ob, manager);

    //Downcast should always work!
    const RawScanResult& raw = dynamic_cast<const RawScanResult&>(ob);
    helper.set(raw);

    //Now publish data - we publish using the ROOTHIST format.
    TH2D* link0 = helper.getScanData(0);
    TH2D* link1 = helper.getScanData(1);
    out << (link0->GetNbinsX()+2)*(link0->GetNbinsY()+2);	//Note, size is number in 1 link ie no bins * no points
    out << SR_DT_ROOTHIST << SR_WD_64;
    
    out.put ( link0->GetArray(), (link0->GetNbinsX()+2)*(link0->GetNbinsY()+2) );
    out.put ( link1->GetArray(), (link1->GetNbinsX()+2)*(link1->GetNbinsY()+2) );
}

void RawScanResultStreamer::read(IStream& in, Streamable& ob, const IOManager& manager) const throw(LogicError, IoError) {
    // cout << "RawScanResultIS refreshing" << endl;
    //Let super class do its work
    ScanResultStreamer::read(in, ob, manager);

    //Downcast should always work!
    RawScanResult& raw = dynamic_cast<RawScanResult&>(ob);
    helper.set(raw);

    //cout <<"Parent finished" << endl;
    //cout << raw.getHeader().getRunNumber() << "  " << raw.getHeader().getScanNumber()
    //     << "  " << raw.getHeader().getModuleName() << endl;

    //Create hists
    double* bins = raw.getPoints().getEdges();

    string name = raw.getUniqueID() + "_link0";
    auto_ptr<TH2D> link0 (new TH2D(name.c_str(), name.c_str(), nChannelLink, -0.5, nChannelLink-0.5,
                     raw.getPoints().getNPoints(), bins));

    name = raw.getUniqueID() + "_link1";
    auto_ptr<TH2D> link1 (new TH2D(name.c_str(), name.c_str(), nChannelLink, -0.5, nChannelLink-0.5,
                     raw.getPoints().getNPoints(), bins));

    delete [] bins;
   
    link0->SetEntries(1);
    link1->SetEntries(1);
    
    //From now on link0 and link1 are not valid.
    helper.setScanData(0, link0);
    helper.setScanData(1, link1);    


    //Now refresh and fill the data
    readData(in, raw);
    
    // cout <<"Done" << endl;
}

void RawScanResultStreamer::readData(IStream& in, RawScanResult& raw) const throw(LogicError, IoError) {
    unsigned int size;
    unsigned short type;
    unsigned short width;
    in >> size >> type >> width;

    switch (type) {
    case SR_DT_SLICE:
        readSliceData(size, width, in, raw);
        return;
    case SR_DT_ROOTHIST:
        readRootData(size, width, in, raw);
        return;
    default:
	ostringstream oss;
	oss << "RawScanResultIS::refreshData Unknown data format: " << type << " size was: " << size;
        throw VersionNotSupportedException(oss.str(), __FILE__, __LINE__); /// AJB 02/05/2003
    }
}

void RawScanResultStreamer::readRootData(unsigned int isize, unsigned short width, IStream& in, RawScanResult& raw) const throw(LogicError, IoError) {
    TH2D* link0 = helper.getScanData(0);
    TH2D* link1 = helper.getScanData(1);

    double* data;
    unsigned int size = 0;
    in.get(&data, size);    
    
    unsigned expectedSize=(nChannelLink + 2) * ( raw.getPoints().getNPoints() + 2 );
    if (size != expectedSize) {
	ostringstream oss;
	oss << "RawScanResultIS::refreshRootData A Wrong data size.  Expected: " << expectedSize;
	oss << " Got: " << size;
        throw StreamCorruptedException(oss.str(), __FILE__, __LINE__);
    }
    
//    cout << "Size = " << size << " expect: " << expectedSize << endl;

    //Copy data
    for (unsigned int i=0; i<size; ++i) {
        link0->SetBinContent(i, data[i]);
    }
    delete [] data;

    in.get(&data, size);
    if (size != expectedSize) {
	ostringstream oss;
	oss << "RawScanResultIS::refreshRootData B Wrong data size.  Expected: " << expectedSize;
	oss << " Got: " << size;
        throw StreamCorruptedException(oss.str(), __FILE__, __LINE__);
    }
    for (unsigned int i=0; i<size; ++i) {
        link1->SetBinContent(i, data[i]);
    }
    delete [] data;

}

void RawScanResultStreamer::readSliceData(unsigned int isize, unsigned short width, IStream& in, RawScanResult& raw) const throw(LogicError, IoError) {
    TH2D* link0 = helper.getScanData(0);
    TH2D* link1 = helper.getScanData(1);

    unsigned int size = 0;
    int npoints = raw.getPoints().getNPoints();
    unsigned expectedSize=2048u * npoints;
	
    
    switch (width) {
    case SR_WD_16:{
	unsigned short* data = 0;
	in.get(&data, size);
	
	if (size != expectedSize && size != isize) {
	    ostringstream oss;
	    oss << "RawScanResultIS::refreshRootData B Wrong data size.  Expected: " << expectedSize;
	    oss << " Got: " << size;
	    throw StreamCorruptedException(oss.str(), __FILE__, __LINE__);
	}
	
	for (int j=0; j<npoints; ++j) {
	    for (unsigned int i=0; i<nChannelLink; ++i) {
		//+1 because bin 0 is underflow
		link0->SetBinContent(link0->GetBin(i+1,j+1), data[j*2048 + i]);
		link1->SetBinContent(link1->GetBin(i+1,j+1), data[j*2048 + i+1024]);
	    }
	}
	delete[] data;	
    }
	return;
	
    case SR_WD_32: {
	UINT32* data = 0;
	in.get(&data, size);
	
	if (size != expectedSize && size != isize) {
	    ostringstream oss;
	    oss << "RawScanResultIS::refreshRootData B Wrong data size.  Expected: " << expectedSize;
	    oss << " Got: " << size;
	    throw StreamCorruptedException(oss.str(), __FILE__, __LINE__);
	}
	
	for (int j=0; j<npoints; ++j) {
	    for (unsigned int i=0; i<nChannelLink; ++i) {
		//+1 because bin 0 is underflow
		link0->SetBinContent(link0->GetBin(i+1,j+1), data[j*2048 + i]);
		link1->SetBinContent(link1->GetBin(i+1,j+1), data[j*2048 + i+1024]);
	    }
	}
	delete[] data;	
    }
	return;
    }
    

}

}
}
