#ifndef MODULEDEFECT_H
#define MODULEDEFECT_H

#include "Sct/Streamable.h"
#include <cstdio>
#include <string>
#include <memory>
#include <map>
#include "Sct/LogicErrors.h"
#include "Sct/SctParameters.h"
#include "ModuleElement.h"

using namespace std;
using Sct::Streamable;

namespace SctData {

/**
This class represents a possible defect with a module.
A ModuleDefect has a description and can be classified as severe or not severe.
@todo Is this enough?  Do we need to add more classes here?  Should we be able to add
other ModuleDefect s anywhere in our code or is having a central repository here enough?
Should this class be renamed ChannelDefect?	
@author Matthew Palmer
*/
    class ModuleDefect : public Streamable {
      public:
	//Defect Prototypes: 
	///@todo add all the different types of defect and check the descriptions
      
        //@{ @name Fit Defects
	static const ModuleDefect DEAD;	        ///< Output always v small
	static const ModuleDefect STUCKON;	///< Output always v high
	static const ModuleDefect OVER; 	///< Occupancy larger than max
	static const ModuleDefect UNDER;	///< Occupancy never reaches mas
        static const ModuleDefect NOINIT;	///< Couldn't initialize the fit.
	static const ModuleDefect FIT_UNDER;
	static const ModuleDefect FIT_OVER;
        static const ModuleDefect SIG_UNDER;
	static const ModuleDefect SIG_OVER;
        //@} 
	//@{ @name Strobe Delay Defects
	static const ModuleDefect SD_LO;	///< Low edge is not between 0 and 35
	static const ModuleDefect SD_HI;	///< Low edge is above 63
	//@}
        //@{ @name NPt Gain Defects	
	static const ModuleDefect LO_GAIN;   ///< Gain < 0.75 * chip average
	static const ModuleDefect HI_GAIN;    ///< Gain > 1.25 * chip average
	static const ModuleDefect LO_OFFSET; ///< Offset < -100
	static const ModuleDefect HI_OFFSET;  ///< Offset > 120
	static const ModuleDefect UNBONDED;     ///< Noise <= 750
        static const ModuleDefect PARTBONDED;   ///< Noise <= 1100
	static const ModuleDefect NOISY;   ///< Noise > 1.15* av chip noise
        //@}
	//@{ @name Trim Defects
	static const ModuleDefect TR_RANGE;      ///< Unexpected chip trim step size
	static const ModuleDefect TR_STEP;       ///< Channel step different from chip
	static const ModuleDefect TR_OFFSET;     ///< Channel offset different from chip
        static const ModuleDefect TR_NOTRIM;    ///< Untrimmable channel
	//@}
        //@{ Other defects
	static const ModuleDefect NO_HI;    /// high noise occupancy
	static const ModuleDefect MEAN_ERROR;
	static const ModuleDefect SIG_ERROR;
	static const ModuleDefect STUCK_CELL; ///< pipeline
	static const ModuleDefect DEAD_CELL; ///< pipeline
	static const ModuleDefect TOKEN;     ///< full bypass token
	static const ModuleDefect RTOKEN;    ///< full bypass token
	static const ModuleDefect TW_HI;     ///< time walk too big
	static const ModuleDefect TW_LO;     ///< time walk too small
	//@}

	//Methods
	/**
	Public constructor: Copy the given ModuleDefect and give ModuleElement
	*/
	ModuleDefect(const ModuleDefect & prototype, const ModuleElement& element) throw () ;
	/**
	Public destructor.
	@note You should never delete a prototype.
	*/
	virtual ~ModuleDefect() throw() {
	}
        //Operations        
        /**
	Return true if this ModuleDefect is severe.
	*/ 
        bool isSevere() const throw() ;
	/** 
	Gets the module element affected by this defect.
	*/
	const ModuleElement& getModuleElement() const throw(LogicError); 
	/**
	True if this is a prototype, false if it actually represents a defect. @todo is this necessary?
	*/
	bool isPrototype() const throw() ;

	/**
	True if this defect applies to a whole chip
	*/
	bool isChipDefect() const throw() ;

	/**
	True if this defect applies to a whole link
	*/
	bool isLinkDefect() const throw() ;

	/**
	* returns true if defect is serious enough to prevent fitting to this channel.
	*/
	bool isUnfittable() const throw() ;

	/**
	* SCTDAQ definitition of unusable is dead or stuckon.
	*/
	bool isUnuseable() const throw() ;

	/**
	* SCTDAQ definitition of `dodgy' is any defect other than dead or stuckon.
	*/
	bool isDodgy() const throw() ;

        /**
        Gets the name of this defect
        */
        const string& getName() const throw() ;

	/**
	 Get a human readable description for this ModuleDefect
	*/
	const string & getDescription() const throw() ;
        
        /**
         Gets a parameter that can be used to determine if something is defective.
         */
        double getParameter() const throw() ;

	/**
	Comparison operators 
	@{  Returns true if the defect types are the same.
	AND defects refer to the same channels.
	*/
	bool isOfTheSameTypeAs(const ModuleDefect & defect) const throw() ;
	/**
	Returns true if the defect types are the same.
	@note defects need NOT refer to the same channels.
	@throw LogicError if LHS or RHS is a prototype 
	*/
	bool operator ==(const ModuleDefect & defect) const throw(LogicError) ;
	
	//Streamable override
	virtual string getClassName() const throw() ;

      private:
	/**
	Private constructor
	*/
	ModuleDefect(int id, bool severe, const string& name, const string & d, double parameter) throw() ;
        /**
        Private constructor.
        Differs from above by assuming that description = sprintf(temp, d, parameter);
        */
        ModuleDefect(int id, bool severe, const string& name, const char* d, double parameter) throw() ;

	/**
	Private copy constructor
	*/
	//ModuleDefect(const ModuleDefect & copy);

	//Data          
	int id;			///< The internal id
	bool severe;		///< True if severe
        string name;            ///< Name of the defect
	string description;	///< The human readable description
        double parameter;       ///< A parameter for deciding if something is defective.
	auto_ptr<ModuleElement> element;	///< The start of the defect

	//IOHelper stuff
	friend class ModuleDefectIOHelper;
	static map <int, ModuleDefect*> defectMap;
    };

//Inlines
    inline ModuleDefect::ModuleDefect(int id, bool severe, const string& name, const string & d, double parameter) throw()
                                      :id(id), severe(severe), name(name), description(d), parameter(parameter), element(0) {
	//Add to map
	defectMap[id] = this;
    }
    
    inline ModuleDefect::ModuleDefect(int id, bool severe, const string& name, const char* d, double parameter) throw() 
                                      :id(id), severe(severe), name(name), description(""), parameter(parameter), element(0) {
	//Add to map
	defectMap[id] = this;

        ///@todo rethink this
        char temp[100];
        sprintf(temp, d, parameter);
        description = temp;                                                 
    }

    inline ModuleDefect::ModuleDefect(const ModuleDefect & prototype, const ModuleElement& el) throw()
	:id(prototype.id), severe(prototype.severe), name(prototype.name), description(prototype.description), parameter(prototype.parameter), element(new ModuleElement(el)) {
    }

    inline double ModuleDefect::getParameter() const throw() {
        return parameter;
    }
    
    inline bool ModuleDefect::isPrototype() const throw() {
	return (element.get()==0);
    } 

    inline bool ModuleDefect::operator ==(const ModuleDefect & defect) const throw(LogicError) {
	return (id == defect.id && getModuleElement() == defect.getModuleElement() );
    }
    inline bool ModuleDefect::isOfTheSameTypeAs(const ModuleDefect & defect) const throw() {
	return (id == defect.id);
    }
    inline const ModuleElement& ModuleDefect::getModuleElement() const throw(LogicError) {
	if (element.get()==0){
	    throw IllegalStateError("ModuleDefect::getModuleElement not available for prototype", __FILE__, __LINE__);
	}
	return *element;
    }
    
    inline bool ModuleDefect::isChipDefect() const throw() {
	return getModuleElement().isChip();
    }
    
    inline bool ModuleDefect::isLinkDefect() const throw() {
	return getModuleElement().isLink();	
    }    
    
    inline bool ModuleDefect::isSevere() const throw() {
	return severe;
    } 
    
    inline bool ModuleDefect::isUnfittable() const throw() {
	return (isOfTheSameTypeAs(OVER) || isOfTheSameTypeAs(UNDER) || 
		isOfTheSameTypeAs(STUCKON) );
    } 
    
    inline bool ModuleDefect::isUnuseable() const throw() {
	return (isOfTheSameTypeAs(DEAD) || isOfTheSameTypeAs(STUCKON) );
    } 
    
    inline bool ModuleDefect::isDodgy() const throw() {
	return !(isOfTheSameTypeAs(DEAD) || isOfTheSameTypeAs(STUCKON) );
    } 
    
    inline const string& ModuleDefect::getName() const throw() { 
        return name;
    }
    
    inline const string & ModuleDefect::getDescription() const throw() {
	return description;
    } 
    
    inline string ModuleDefect::getClassName() const throw() {
	return "SctData::ModuleDefect";
    }
}
#endif				//#ifndef MODULEDEFECT_H
