#ifndef MODULEDEFECTLIST_H
#define MODULEDEFECTLIST_H

#include "Sct/Streamable.h"
#include "Sct/Exception.h"
#include "ModuleElement.h"
#include <vector>
#include <memory>
#include <boost/shared_ptr.hpp>

using namespace std;
using namespace Sct;
using boost::shared_ptr;

namespace SctData {
	
class ModuleDefect;
class ModuleElement;
	
/**
This class represents all the defects in a module.
It is essentially a map of channel number to a collection of ModuleDefect s.
@author Matthew Palmer
*/
class ModuleDefectList : public virtual Streamable {
public:
	/**
	Create a new empty ModuleDefects object.
	*/
	ModuleDefectList() throw() ;
	
	/**
	Destructor deletes all ModuleDefect objects as we own them
	*/
	virtual ~ModuleDefectList() throw() ;

    /**
	Add a defect. 
	@note Takes a prototype rather than a precreated ModuleDefect to make sure we own them,
	because we delete them in the destructor.
	@param prototype The prototype ModuleDefect to add
	@param ModuleElement, i.e. channel/chip/link/module
    */
    
        void addDefect(const ModuleDefect& prototype, const ModuleElement& element) throw() ; 
    /**
       Get all defects in the ModuleDefectList.
       @return A reference to a list of all defects.
    */
    const vector<shared_ptr<const ModuleDefect> >& getAllDefects() const throw() ;

    /**
       Methods enquiring as to defects on ANY PART of a module element
       @{
       Get all the defects affecting a given channel.
       @param channelNum The channel number
       @return a vector of const ModuleDefect*
    */
    shared_ptr<ModuleDefectList> getDefectsAffectingElement(const ModuleElement&) const throw(LogicError);	
    int getNDefectsAffectingElement(const ModuleElement&) const throw(LogicError) ;
    /// check for severe defects.
    bool severeDefectAffectingElement(const ModuleElement&) const throw(LogicError) ;
    /// check for any defects.
    bool defectAffectingElement(const ModuleElement&) const throw(LogicError) ; 
    /// check for defects of a particular type.
    bool defectAffectingElement(const ModuleElement&, const ModuleDefect&) const throw(LogicError) ;

    
    //@}
    //@{
    //@name Methods enquiring as to defects which affect the WHOLE of an element.
    /// get all defects who contain the element
    shared_ptr<ModuleDefectList> getDefectsEncompassingElement(const ModuleElement&) const throw(LogicError);
    /// number of defects inside element.
    int getNDefectsEncompassingElement(const ModuleElement&) const throw(LogicError) ;
    /// check for severe defects.
    bool severeDefectEncompassingElement(const ModuleElement&) const throw(LogicError) ;
    /// check for any defects.
    bool defectEncompassingElement(const ModuleElement&) const throw(LogicError) ; 
    /// check for defects of a particular type.
    bool defectEncompassingElement(const ModuleElement&, const ModuleDefect&) const throw(LogicError) ; 
    //@}
    ///Streamable override
    virtual string getClassName() const throw();

    ///Add defects from another list.
    ModuleDefectList& operator +=(const ModuleDefectList&) throw();
    
    /// print the defects:
    ostream& print(ostream& out) const throw();

private:
    vector< shared_ptr<const ModuleDefect> > defectList;		///< All defects
    /// allow this class to share its defects with ModuleDefectLists it creates
    void add( shared_ptr<const ModuleDefect> ) throw();
};
	
}

#endif //#ifndef MODULEDEFECTLIST_H
