#ifndef TESTRESULT_H
#define TESTRESULT_H

#include "ModuleDefectList.h"
#include "Sct/Serializable.h"
#include "Sct/Exception.h"
#include "ModuleConfiguration.h"
#include <string>
#include <vector>
#include <boost/shared_ptr.hpp>
#include "Sct/LogicErrors.h"
#include "Sct/IoExceptions.h"

using namespace std;
using namespace Sct;
using boost::shared_ptr;

namespace SctData {
class FitScanResult;
class RawScanResult;
class ConfigurationVariable;

/**
This class is the Abstract Base Class for all TestResults.  Examples are Response Curves, Trim Range Tests and StrobeDelay Test.
It provides useful data and functions to the sub-classes.  For instance, identifying information such as the list of 
scan numbers, the run number, module name and the test points.
@todo Think: Do we want a TestHeader class?  Such that a Test is a collection of TestHeaders
@author Matthew Palmer
@date 12 March 2003
*/
class TestResult:public virtual Serializable {
public:
    /// Constructor is in protected
    ~TestResult() throw() {};
    
    ///Serializable override
    ///@throw LogicError if TestResult is empty.
    virtual string getUniqueID() const throw(LogicError) ;    
    
    ///@return the run number
    unsigned int getRunNumber() const throw() {
        return runNumber;
    }

    void setRunNumber(unsigned int runNumber) throw();

    ///@return the module name
    string getModuleName() const throw() {
        return moduleName;
    }

    void setModuleName(string name) throw();

    /**
      Get the configuration variable associated with all the scans in this test.
      */
    const ConfigurationVariable& getScanVariable() const throw();

    void setScanVariable(const ConfigurationVariable& scanVariable) throw();

    /**
      Get the configuration variable associated with this test.
      */
    const ConfigurationVariable& getTestVariable() const throw();

    void setTestVariable(const ConfigurationVariable& testVariable) throw();

    /**
      Get the updated configuration.
      */
    const ModuleConfiguration& getUpdatedConfiguration() const throw();
    ModuleConfiguration& getUpdatedConfiguration()  throw();

    /// gets the scan number at a particular postition `i' within the data.
    unsigned getScanNumberAt(unsigned int i) const throw(LogicError);

    /// gets the test point at a particular position `i' within the data.
    double getTestPointAt(unsigned int i) const throw(LogicError);

    /**
    Get the defects found in the TestResult.
    @return The ModuleDefects class representing all the defects found in this TestResult.
    */
    const ModuleDefectList& getDefects() const throw() ;
    ModuleDefectList& getDefects() throw() ;

    /**
      Return true if the module passed the Test
      */
    bool getPassed() const throw();
    
    /**
      Set whether the module passed the scan
      */
    void setPassed(bool passed) throw();
    
    /**
      Return true if the module had a problem
      */
    bool getProblem() const throw();
    
    /**
      Set whether the module had a problem
      */
    void setProblem(bool problem) throw();
    
    /**
      Return all the comments.
      */
    vector<string> getComments() const throw(); 
    
    /**      
      Add a comment - must be a single line (no newlines)
      */
    void addComment(string comment) throw();
    
    /// Add a scan to this test.
    virtual void addScan(const unsigned int scanNumber, const double testPoint) throw(LogicError);


    /**
    Returns an iterator that is a model of (at least) const forward iterator that can be used to
    traverse all the scanNumbers.
    */
    //ScanNumberIterator& getScanNumbers();

    /**
     * @return the number of scans in this test.
     */
    const unsigned int getNScans() const throw() {
        return data.size();
    }

    /**
    * Find the index for a particular scanNumber.
    * @throw LogicError if not found.
    */
    unsigned int getIndex(const unsigned int scanNumber) const throw(LogicError) ;

    /**
    * Gets a FitScanResult that has previously been added with addFit  
    * @return A pointer to a FitScanResult corresponding to the scanNumber given.
    * @throw InvalidArgument if scan is not in this TestResult
    * @throw IoError if fit can't be obtained
    */
    shared_ptr<const FitScanResult> getFit(unsigned int index) const throw(LogicError,IoError) ;

    /**
    * Gets a RawScanResult that has previously been added with addRaw.
    * @return A pointer to a RawScanResult corresponding to the scanNumber given.
    * @throw InvalidArgument if scan is not in this TestResult
    * @throw IoError if fit can't be obtained
    */
    shared_ptr<const RawScanResult> getRaw(unsigned int index) const throw(LogicError,IoError) ;

    /**
	Add a FitScanResult. Puts it in the correct place.
    */
    void addFit(shared_ptr<const FitScanResult> fit) throw(LogicError) ;
    /**
       Add a RawScanResult. Puts it in the correct place.
    */
    void addRaw(shared_ptr<const RawScanResult> raw) throw(LogicError) ;
    /**
       convenience method checks if all fits are in place.
       @return true if none of the fits are null pointers.
    */
    bool hasAllFits() const throw();
     /**
       convenience method checks if all raws are in place.
       @return true if none of the raws are null pointers.
    */
    bool hasAllRaws() const throw();
protected:
    /**
      Default constructor.
      */
    TestResult() throw();

    /**
     * Create a TestResult from all the stuff needed to initialize it.
     */
    TestResult(const unsigned int runNumber, const string & moduleName, const ConfigurationVariable& testVariable,
	       const ConfigurationVariable& scanVariable) throw ();

    /**
    * small internal class contains all the data relevant to a particular scan.
    */
    class ScanData {
    public:

        /**
	   constructor of ScanData 
	   @param scanNumber : the scan number 
	   @param testPoint the value of the variable changed between scans.
	   @param raw a shared pointer to the appropriate RawScanResult
	   @param fit a shared pointer to the appropriate FitScanResult
        */
        ScanData(const unsigned int scanNumber, const double testPoint, shared_ptr<const RawScanResult> raw, 
		 shared_ptr<const FitScanResult> fit) throw();
	
	/**
	   minimal constructor
	   @param scanNumber : the scan number 
	   @param testPoint the value of the variable changed between scans
	*/
	ScanData(const unsigned int scanNumber, const double testPoint) throw();
        //@{
        /// Operators needed for sorting
        bool operator<(const ScanData & s) const throw();
        bool operator>(const ScanData & s) const throw();
        bool operator==(const unsigned int aScanNumber) const throw();
        //@}

        unsigned int scanNumber;	///< The scan number
        double testPoint;	///< The value of the variable changed between scans
        mutable shared_ptr<const RawScanResult> raw;	///< Cache of the appropriate RawScanResult
        mutable shared_ptr<const FitScanResult> fit;	///< Cache of the appropriate FitScanResult
    };

    vector <ScanData> data;	/// vector of scanData (internal to this class) objects: one for each scan in the Test.
    unsigned int runNumber;	///<The run number
    string moduleName;	///< The module name.

    const ConfigurationVariable* testVariable;		///< The test variable
    const ConfigurationVariable* scanVariable;		///< The scan variable for all the ScanNumbers in this TestResult
    ModuleConfiguration updatedConfig;			///< The updated configuration
    ModuleDefectList defects;	                        ///< The ModuleDefects found in producing this TestResult.
    bool passed;					///< True if the module passed the Test
    bool problem;					///< True if there was a problem with the Module
    vector<string> comments;				///< Comments

    //IOHelper stuff
    friend class TestResultIOHelper;
};

}				// end of namespace SctData
#endif				//TESTRESULT_H
