// File: SctApi.h
// Library interface to SCTAPI
#ifndef SCTROD_SCTAPI_H
#define SCTROD_SCTAPI_H

#include <list>
#include <map>
#include <string>
#include <vector>

#include <boost/thread.hpp>
#include <boost/shared_ptr.hpp>

#include <mrs/message.h>
#if USE_IS
#include <is/isinfotmpl.h>
#endif

#include "sctddc/SCTDCSCommand.hxx"
#include "log.h"

// This defines ABCDModule (a typedef)
#include <CommonWithDsp/sctStructure.h>
// Unfortunately this is necessary
#include <CommonWithDsp/processor.h>

#warning "Should some of these return smart pointers?"

struct scan_result_ptrs;

namespace SctConfiguration {
  class Configuration;
}

namespace SctApi {
class Crate;

class PrimListWrapper;

/// Trigger interface
class Trigger {
 public:
  virtual ~Trigger();

  /// Type for representation as a series of pairs as defined for ROD DSP
  typedef std::vector<std::pair<UINT16, UINT32> > RODTriggers;

  /// Set to a single L1A
  virtual void singleL1A() = 0;
  /// Set to a double L1A
  virtual void doubleL1A(unsigned short delay) = 0;
  /// Set to a Calibration pulse followed by an L1A BC0s later
  virtual void calL1A(unsigned short delay) = 0;
  virtual void pulseL1A(unsigned short delay) = 0;
  virtual void softL1A(unsigned short delay) = 0;
  virtual void softCalL1A(unsigned short delay, unsigned short delay2) = 0;
  virtual void softPulseL1A(unsigned short delay, unsigned short delay2) = 0;
  virtual void bcL1A(unsigned short delay) = 0;
  virtual void bcCalL1A(unsigned short delay, unsigned short delay2) = 0;
  virtual void bcPulseL1A(unsigned short delay, unsigned short delay2) = 0;

  virtual void setCommIncr(unsigned short command, unsigned short incr) = 0;
  virtual void getCommIncr(unsigned short &command, unsigned short &incr) const = 0;

  /// Get list of triggers in ROD format
  virtual const RODTriggers getRODTriggers() const = 0; 

  virtual void print() const = 0;
};

/**
    Scan class

    Describes a scan request

    There are two module sets defined by the module group 
    number in the configuration.
*/ 
class Scan {
 public:
  virtual ~Scan();

  /// Typedef for list of scan points
  typedef std::vector<FLOAT32> ScanPoints;

  /// Current options for a scan request
  enum ScanOptions {FORMAT, FULL, BITS32, LOOPCALLINE, DISTSLAVE, DEBUG};

  virtual void configure(unsigned short type, float start, float stop, float step) = 0;
  virtual void configure2(unsigned short type, float start, float stop, float step) = 0;

  virtual const Trigger *getTrigger1() const = 0;
  virtual const Trigger *getTrigger2() const = 0;

  virtual Trigger *getTrigger1() = 0;
  virtual Trigger *getTrigger2() = 0;

  virtual void setTrigger1(Trigger *trigger) = 0;
  virtual void setTrigger2(Trigger *trigger) = 0;

  virtual void setScanVariable1(unsigned short var) = 0;
  virtual void setScanVariable2(unsigned short var) = 0;

  virtual unsigned short getScanVariable1() const = 0;
  virtual unsigned short getScanVariable2() const = 0;

  virtual void setNTrigs(unsigned long nTrigs) = 0;
  virtual unsigned short getNTrigs() const = 0;

  virtual const ScanPoints getScanPoints1() const = 0;
  virtual const ScanPoints getScanPoints2() const = 0;
  virtual void setScanPoints1(const ScanPoints &scans) = 0;
  virtual void setScanPoints2(const ScanPoints &scans) = 0;

  virtual int getOption(enum ScanOptions) const = 0;
  virtual void setOption(enum ScanOptions, int option) = 0;

  /// Print the scan
  virtual void print() const = 0;

  virtual void setScanNumber(unsigned int scan) = 0;
  virtual unsigned int getScanNumber() const = 0;
  virtual void setRunNumber(unsigned int scan) = 0;
  virtual unsigned int getRunNumber() const = 0;

  virtual std::list<std::string> getModuleList(unsigned int group) const = 0;
  virtual void setModuleList(unsigned int group, std::list<std::string> newList) = 0;
  virtual unsigned int getNGroups() const = 0;
  virtual void setNGroups(unsigned int val) = 0;
};

/// Internal class with lots more information
/**
   All this stuff restricts a scan to one ROD
 */
class ScanEx {
 public:
  // Which ROD we're going to talk to (filled in automatically)
  int partition;
  int crate; 
  int rod;
  UINT32 channels[2];

  UINT32 dspChannels[4][2];

  /// Number of slaves
  int slaves;

  /// Which DSPs to use
  char bitFieldDSP;

  /// What the diagnostic reg was set to... for debug mode
  unsigned int diagnosticReg;

  /// Is this a trim scan?
  int trimScan;

  //    *************** Unused but useful fields

  /// Trigger source
  int timTrigs;
};

/** 
    SctApi class.

    This is the API used by the rest of the world to interact with
    SCT Modules using RODs.

    Functionality that should be included in this class includes.


    DAQ run level transitions:
       Boot
           Load and run the crate controller (Done automatically in DAQ-1 from configuration)
       Load
           Load config?
       Config
           Send config to Modules
       Start
           Mode for execution of scans

    Configuration:
      loading from file
      loading to modules
      fine grained control on modules
      recovering module configuration from the ROD
      Save configuration

      Reconfiguring from new configuration file
    
    Scanning:
      perform scans over configuration variable(s)

    DCS:
      minimal DCS communication

    ROD diagnostics:
      Flash LEDs
      Text buffer retrieval
      Memory dumps
      Send arbitrary primitives to a single ROD

    Indexing:
      There are several ways of addressing a ROD or module
      the canonical method is
              partition, crate, rod, module

    Other methods include:
       Geographical:
           barrel/disc MUR, number
       MUR:
           Partition, MUR, number
*/
class SctApi {
  SctApi(const SctApi &);
  SctApi &operator=(const SctApi&);
 public:

  /** The constant returned as an unknown module */ 
  static const unsigned int BAD_MODULE;

  /**
     Default to local xmlConfig
  */
  SctApi();

  /**
     Specify configuration
  */
  SctApi(boost::shared_ptr<SctConfiguration::Configuration> newConf);

  /**
     Specify initial debug level
  */
  SctApi(int initDebug);

  ~SctApi();

 private:
  /**
     This sets the unexpected handler to interpret all sorts of exceptions

     Called by constructor
   */
  void setup();

public:
  /**
     C++ only method to tell a MRS aware program where to write to
   */
  void setMrsStream(MRSStream *stream);

#if USE_IS
  /**
     C++ only method to tell a IS aware program where to write to
   */
  void setIsDictionary(ISInfoDictionary *dict);
#endif

  /***************** Global configuration functions ***********************/

  /// Initialise all software components
  /** Initialise the crates and create the Rod/Vme objects.

      Load the system configuration from the database.
      Create crates and RODs as necessary.
      Calls initialiseRod and initialiseModule as necessary.

      Store the module configuration in the RODs (UNIMPLEMENTED)

      Return: 0 for success
              Possibly error code for failure

      IPC: Block everything else!
  */
  void initialiseAll(int run);

  /// Tidy up and shutdown
  /** Leave everything in a suitable state and delete all the objects

      Shutdown anything that needs shutting down
      This is where we should send DCS to standby message??

      IPC: Block everything else!
  */
  void shutdownAll();

  /// Change the run number
  /**
     IPC: Don't do during a scan
  */
  void setRunNumber(UINT32 newRun);

  /// Change the next scan number
  /**
     IPC: Don't do during a scan
  */
  void setScanNumber(UINT32 newScan);

  /// Send the configurations from the RODs to the modules.
  /**
     DDC: Check modules in configurable state.
          Correct run state and no warnings

     IPC:  Don't do during a scan
     Conf: Maybe interaction
   */
  void configureAllModules();

  /*********************** Module configuration **********************/

  /// (Re)Load configuration from the "configuration database"
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void loadConfiguration();

  /// Load new configuration from a "configuration database"
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void loadConfiguration(std::string filename);

  /// Store the current module configurations into database
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void storeModuleConfigurations();

  /// Reset BOC configuration from current 
  /**
     IPC:  Don't do during a scan
     Conf: Definite interaction!
  */
  void configureBOC(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Wait for a ROD to finish a primlist
  /**
     IPC:  Allow during a scan (only expert mode)
     Conf: no interaction!
  */
  void awaitResponse(unsigned int partition, unsigned int crate, unsigned int rod, int timeout);

  /// Wait for all RODs in a crate to finish a primlist
  /**
     IPC:  Allow during a scan (only expert mode)
     Conf: no interaction!
  */
  void awaitResponseAll(unsigned int partition, unsigned int crate, int timeout);

  /// Return the out list
  /**
     (No interaction with VME)
     IPC:  Allow during a scan (only expert mode)
     Conf: No interaction!
  */
  unsigned long *getResponse(unsigned int partition, unsigned int crate, unsigned int rod, 
                             unsigned long &length);

  /** Get a message in the text buffer

      Get a message from a ROD text buffer
      Returns whether the action was completed
  */
  bool getRodMessage(unsigned int partition, unsigned int crate, unsigned int rod,
                     char *buffer, int &length);

  /*************** ROD Diagnostics ********************/

  /// Flash the LED associated with a slave DSP. 
  /**
     This creates its own primitive list and sends it to 
     a slave DSP on the ROD.

     Period given in ms

     IPC:  Possible during a scan?
     Conf: No interaction
  */
  void flashLED(unsigned int partition, unsigned int crate, unsigned int rod,
                unsigned long slaveNumber, long period = 1000, long flashes = 10);

  /// Echo data to a ROD
  /**
     IPC:  Possible during a scan?
     Conf: No interaction
  */
  void echo(unsigned int partition, unsigned int crate, unsigned int rod,
            unsigned int length, const long *data);


  /***************** Direct ROD access (Debug only) ********************/
  /* Should be able to do anything with these commands */

  /// Dump a block of memory direct from a DSP
  /**
     Dump contents of a block of DSP memory to cout.
     Use dspNumber -1 to refer to the master DSP.

     IPC:  Possible during a scan
     Conf: No interaction
  */
  int dspBlockDump(unsigned int partition, unsigned int crate, unsigned int rod,
                   long dspStart, long numWords, long dspNumber);

  int dspBlockDumpFile(unsigned int partition, unsigned int crate, unsigned int rod,
                       long dspStart, long numWords, long dspNumber, string filename);

  /// Dump registers to files
  void standardRegisterDump(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Read a block of memory direct from a DSP
  /**
     Use dspNumber -1 to refer to the master DSP.

     IPC:  Possible during a scan
     Conf: No interaction
  */
  unsigned long *dspBlockRead(unsigned int partition, unsigned int crate, unsigned int rod,
                              long dspStart, long numWords, long dspNumber, long &length);

  /// Write a block direct to a DSP
  /**
     Use dspNumber -1 to refer to the master DSP.

     IPC:  Possible during a scan
     Conf: No interaction
  */
  int dspBlockWrite(unsigned int partition, unsigned int crate, unsigned int rod,
                    unsigned long *buffer, unsigned long dspAddress, long numWords, long dspNumber);

  ///   Read a block of slave DSP memory using RW_SLAVE_MEMORY
  /**
     Don't use if important to preserve primitive state!

     Returns new[]ed block of memory

     IPC:  Possible during a scan
     Conf: No interaction
  */
  unsigned long *readSlaveDsp(unsigned int partition, unsigned int crate, unsigned int rod,
                          int s, int add, int words);

  /// Read one ROD Status Register via HPI
  /**
     IPC:  Immediate response
     Conf: No interaction
  */
  unsigned long readRodStatusReg(unsigned int partition, unsigned int crate, unsigned int rod,
                                 long regNumber);

  /// Read one ROD Command Register via HPI
  /**
     IPC:  Immediate response
     Conf: No interaction
  */
  unsigned long readRodCommandReg(unsigned int partition, unsigned int crate, unsigned int rod,
                                  long regNumber);

  /// Read a single 32b word from MasterDSP SDRAM via HPI
  /**
     IPC:  Immediate response
     Conf: No interaction
  */
  unsigned long dspSingleRead(unsigned int partition, unsigned int crate, unsigned int rod,
                              const unsigned long dspAddr, long dspNumber);

  /// Write a single 32b word to MasterDSP SDRAM via HPI
  /**
     IPC:  Immediate response (expert only)
     Conf: No interaction
  */
  void dspSingleWrite(unsigned int partition, unsigned int crate, unsigned int rod,
                      unsigned long dspAddr, unsigned long buffer, long dspNumber);

  /// Retrieve module configuration data
  /**
     Module data as cached from Configuration

     IPC:  Immediate response
     Conf: Interaction
  */
  ABCDModule *retrieveModule(UINT32 mid);

  /// Create a prim list
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void createDebugPrimList();
  /// Insert a primitive in a prim list
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void addDebugPrimList(unsigned long length, long index, long id, long version,
                        unsigned long * body);
  /// Send the previously created prim list
  /**
     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void sendDebugPrimList(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Send the "Debug" primitive to a slave dsp
  /**
     Send the previously created prim list to a slave dsp
     @param await Master prim list waits for slave list to complete 
     @param response Expect a reply 

     IPC:  Immediate response (expert)
     Conf: No Interaction
  */
  void sendDebugSlavePrimList(unsigned int partition, unsigned int crate, unsigned int rod,
                              unsigned int slave, bool await, bool response);


  /**** Official SCTAPI bit (majority unimplemented) ******/

  /// Return unique identifier for a module
  /**
     @param sn Module serial number
     @return The unique identifier, BAD_MODULE if not found

     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findModule(std::string sn);

  /// Return string for uid
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  std::string convertToString(UINT32 mid);

  /// Return unique identifier for mur/module 
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findModule(INT32 mur, INT32 module);

  /// Return mur/module pair for uid
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  std::pair<INT32, INT32> convertToMUR(UINT32 mid);

  /// Return unique identifer from barrel geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findBarrelModule(INT32 barrel, INT32 row, INT32 number);

  /// Convert uid to barrel geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  void convertToBarrelModule(UINT32 mid, UINT32 &barrel, UINT32 &row, int &number);

  /// Return unique identifer from endcap geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  UINT32 findEndcapModule(INT32 disk, INT32 quadrant, INT32 number);

  /// Convert uid to endcap geometry
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  void convertToEndcapModule(UINT32 mid, UINT32 &disk, UINT32 &quadrant, UINT32 &number);

  /// Get all modules from ROD to memory
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  void getABCDModule(UINT32 bank);

  /// Get module from ROD to memory
  /**
     IPC:  Immediate response
     Conf: Lookup
  */
  void getABCDModule(UINT32 mid, UINT32 bank);

  /// Set module config in ROD bank
  /**
     IPC:  Don't do during scan
     Conf: Possible cache
  */
  void setABCDModule(UINT32 mid, UINT32 bank);

  /// Send configuration for all modules from the ROD bank to the module
  /**
     @param bank Configuration bank in the ROD
     @param type How much configuration to transmit

     DDC:  Check module state
           Set select line
     IPC:  Not during scan
     Conf: No Lookup
  */
  void sendAllABCDModules(UINT32 bank, UINT32 type = 2);

  /// Send configuration for one module from the ROD bank to the module
  /**
     @param mid Unique identifier of the module to send
     @param bank Configuration bank in the ROD
     @param type How much configuration to transmit

     DDC:  Check module state
           Set select line
     IPC:  Not during scan
     Conf: No Lookup
  */
  void sendABCDModule(UINT32 mid, UINT32 bank, UINT32 type=2);

  /// Modify the mask of a module configuration
  /**
     Set the channel mask of all chips in a module

     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDMask(UINT32 mid, UINT32* mask);

  /**
     Modify the mask of a list of module configuration
  */
/*    void modifyABCDListMask(mList* mlist, UINT32* mask); */

  /**
     Modify a module's trim configuration

     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDTrims(UINT32 mid, UINT8* trims);

  /**
     Modify a module list's trim configuration
  */
/*    void modifyABCDTrims(mList* mlist, UINT8* trims); */

  /// Modify all module configurations
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVar(UINT32 typ, FLOAT32 var);

  /// Modify variable in module configuration
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVar(UINT32 mid, UINT32 typ, FLOAT32 var);

  /// Modify variable in list of modules
/*    void modifyABCDVar(mList* mlist, UINT32 typ, FLOAT32 var); */

  /// Modify variable on one chip of module configuration 
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDVar(UINT32 mid, UINT32 chip, UINT32 typ, FLOAT32 var);

  /// Modify variable on one chip of module configurations in a list
/*    void modifyABCDVar(mList* mlist, UINT32 chip, UINT32 typ, FLOAT32 var); */

  /// Modify response curve setting on one chip
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyABCDRC(UINT32 mid, UINT32 chip, UINT16 function, FLOAT32 p0, FLOAT32 p1, FLOAT32 p2);

  /// Modify BOC parameter
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void modifyBOCParam(unsigned int partition, unsigned int crate, unsigned int rod,
                      unsigned int channel, unsigned int type, unsigned int val);

  /// Print BOC parameters for all channels
  /**
     IPC:  Immediate response (Needs results by IPC?)
     Conf: No communication
  */
  void printBOCSetup(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Print BOC registers (independent of channels)
  /**
     IPC:  Immediate response (Needs results by IPC?)
     Conf: No communication
  */
  void printBOCRegisters(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Turn all the BOC laser currents off (modules lose clock, not a good idea)
  /**
     IPC:  Immediate response
     Conf: No communication
  */
  void lasersOff();

  /* *************  TIM functs **************/

  /// Start the TIM generating regular triggers 
  /**
     Freq is rounded to the nearest power of 10 between .01 and 100k Hz

     IPC:  Not during scan
     Conf: No communication
  */
  void freeTriggers(unsigned int partition, unsigned int crate,
                    float freq);

  /// Stop the TIM generating regular triggers
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void stopTriggers(unsigned int partition, unsigned int crate);

  /// Tell the TIM to send an L1A 
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void timL1A(unsigned int partition, unsigned int crate);

  /// Tell TIM to send Cal + L1A
  /**
     IPC:  Not during scan
     Conf: No communication
  */
  void timCalL1A(unsigned int partition, unsigned int crate, int delay);


  /// Print status to cout
  /**
     IPC:  Immediate response (sane in all eventualities!)
     Conf: No communication
  */
  void status();

  /// Do a scan (hardcoded for testing)
  /**
     IPC:  Not during any other scan
     Conf: No communication
  */
  void defaultScan(int type);

  /// Coordinate a scan
  /**
     DDC:  Check modules configured...
           Check select
     IPC:  Not during any other scan
     Conf: Lookup
  */
  void doScan(Scan &scan);

  /**
     Abort any long term activity

     DDC:  No action
     IPC:  Immediate response
     Conf: No communication
   */
  void abortScan();

  /// Scan returning histograms of raw data
  /**
     IPC:  Not during any other scan
     Conf: Lookup
  */
  void doRawScan(Scan &scan, int delay, bool configure = false, bool clkBy2 = false);

  /// Send one trigger to all modules
  /**
     IPC:  Not during any other scan
     Conf: No communication
  */
  void sendTriggerAll(const Trigger *trig);

  /// Send one trigger to modules in one ROD
  /**
     IPC:  Not during any other scan
     Conf: No communication
  */
  void sendTrigger(unsigned int partition, unsigned int crate, unsigned int rod, const Trigger *trig);

  /// Print the module configuration in memory
  /** 
     Uses the sctConf method

     IPC:  Not very useful without feedback
     Conf: Possible lookup
  */
  void printABCDModule(int mid);

  /// Print the module configuration from the ROD 
  /**
     First retrieve the configuration from the ROD
     Uses the sctConf method

     IPC:  Not very useful without feedback
     Conf: Possible lookup
   */
  void printABCDRodModule(int mid, int bank);

  /// Print decoded event
  /**
     IPC:  Needs feedback, immediate response
     Conf: No Communication
  */
  void decodeEvent(unsigned int partition, unsigned int crate, unsigned int rod, 
                   int sl, int index, bool extFlag = false, bool errorType = false);

  // Very bad but CINT isn't happy at the moment...
  void printHex(UINT32 num, int width);

  /// Read raw link data (send L1A to trigger some data)
  /**
     IPC:  Needs feedback, immediate response
     Conf: No Communication
  */
  void rawData(unsigned int partition, unsigned int crate, unsigned int rod, int delay, int units, bool setMask = true, const Trigger *trig = 0);

  /// Send an L1A to all modules
  /**
     The mask is unchanged in this function and it is set up to go to all modules

     IPC:  Not during scan
     Conf: No Communication
  */
  void sendL1A(unsigned int partition, unsigned int crate, unsigned int rod, bool capture = false);

  /// Make guess at what's attached to different channels
  /**
     Do quick scan of channels
     Recognises all the same, clock by 2 and an event header
     Return a char for each link examined
        0 All zero's
        1 All one's
        2 Clock by 2
        L L1A
        E Event Header
        S Soft Reset
        B BC Reset
        J Junk (ie anything else)

     IPC:  Needs feedback, some delay to response
     Conf: Some lookup?!
  */
  vector<char> probe(unsigned int partition, unsigned int crate, unsigned int rod, signed int harness = -1);

  /// Send a trigger to all channels and see what comes back
  /**
     Do quick scan of channels
     Recognises all the same, clock by 2 and an event header
     Return a char for each link examined
        0 All zero's
        1 All one's
        2 Clock by 2
        L L1A
        E Event Header
        S Soft Reset
        B BC Reset
        J Junk (ie anything else)

     IPC:  Needs feedback, some delay to response
     Conf: Some lookup?!
  */
  vector<char> probeWithTrigger(unsigned int partition, unsigned int crate, unsigned int rod, const Trigger *trigger, signed int harness = -1);

  /**
     Do a "BOC histogram"
   */
  void bocHistogram(unsigned int partition, unsigned int crate, unsigned int rod, 
                    unsigned int samples, unsigned int numLoops);

  /*** DCS functionality */

  /**
     Request hard reset on all modules

     DDC:  Primary purpose
     IPC:  Not during scan
     Conf: Some lookup?!
  */
  void requestHardResetAll();

  /**
     Request hard reset of module mid

     DDC:  Primary purpose
     IPC:  Not during scan
     Conf: Some lookup?!
   */
  void requestHardReset(UINT32 mid);

  /**
     Request initialisation of an IV scan

     DDC:  Primary purpose
     IPC:  Not during scan, result will be asynchronous
     Conf: Some lookup?!
   */
  void requestIVCurve(UINT32 mid, FLOAT32 start, FLOAT32 stop, FLOAT32 step, UINT16 delay, FLOAT32 currentLim);

  /**
     Request initialisation of an IV scan on all modules

     DDC:  Primary purpose
     IPC:  Not during scan, result will be asynchronous
     Conf: Some lookup?!
   */
  void requestIVCurveAll(FLOAT32 start, FLOAT32 stop, FLOAT32 step, UINT16 delay, FLOAT32 currentLim);

  /** Get a channel parameter from DDC */
  void ddcGetChannelParameter(int crate, int channel, std::string dpeName,
                              int& dpeValue, int timeout);

  /**
     Set the select value for module mid

     This is a debug method for DCS. Primary route to this functionality is 
     via modifyABCDVar and sendModule!

     DDC:  Primary purpose
     IPC:  Not during scan
     Conf: Some lookup?!
   */
  // #warning "Should this just change DCS?"
  void setSelect(UINT32 mid, bool value); 

  /**
     Change run state of all power modules

     One of Off, Standby, On

     DDC:  Primary purpose
     IPC:  Not during scan
     Conf: No communication
  */
  void changeRunState(int state);

  /**
     Stop all tasks to do with histogramming

     This is a lower level function than abort,
     which returns to a know state

     DDC:  If changed?
     IPC:  Any time
     Conf: No communication
   */
  void tidyHistogramming();

  /**
     Set the current debug level

     DDC:  No action 
     IPC:  Class action
     Conf: No communication
  */
  void setDebugLevel(unsigned int level);

  /**
     Get the current debug level

     DDC:  No action 
     IPC:  Class action, immediate response
     Conf: No communication
  */
  unsigned int getDebugLevel();

  void debugStepHistogram();

  void debugContinueHistogram();

  void debugAbortHistogram();

  /// Direct write of ROD mode!
  void rodMode(unsigned int partition, unsigned int crate, unsigned int rod,
               int mode, int flag, int fifoSetup, int nBins, int delay, int message);

  /// Set module masks
  void setupModuleMask(unsigned int partition, unsigned int crate, unsigned int rod,
                       int port, int slvs);

  /// Add set module masks command to primitive list
  void setupModuleMask(int port, int slvs, boost::shared_ptr<PrimListWrapper> list);

  void sendData(unsigned int partition, unsigned int crate, unsigned int rod);
 private:
  /********** Should these be part of the API? **************/

  ///Initialise a ROD from the configuration
  /**
      Proxy for Crate

      Return 0 for success
  */
  int initialiseRod(unsigned int partition, unsigned int crate, unsigned int rod);

  ///Initialise a TIM from the configuration
  /**
      Proxy for Crate

      Return 0 for success
  */
  int initialiseTim(unsigned int partition, unsigned int crate);

  /// Initialise a BOC from the configuration
  /**
      Proxy for Crate

      Return 0 for success
  */
  int initialiseBOC(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Write block of memory to a slave (using RW_SLAVE_MEMORY)
  void writeSlaveBlock(unsigned int partition, unsigned int crate, unsigned int rod, int s, long sAdd, int words, long *block);

  /// Set block of slave memory to value (using RW_SLAVE_MEMORY)
  void setSlaveBlock(unsigned int partition, unsigned int crate, unsigned int rod, int s, long sAdd, int words, long value = 0);

  /// Write to a ROD register
  void writeRODRegister(unsigned int partition, unsigned int crate, unsigned int rod,
                        int reg, int off, int width, int value);

  /// Add write rod register command to primitive list
  void writeRODRegister(int reg, int off, int width, int value, boost::shared_ptr<PrimListWrapper> list);

  /// Read a register from the ROD
  unsigned int readRODRegister(unsigned int partition, unsigned int crate, unsigned int rod, int r);

  /// Poll a register
  int pollRegister(unsigned int partition, unsigned int crate, unsigned int rod, int r, int off, int width, int val, int timeout = 1000000);

  unsigned long *readFifo(unsigned int partition, unsigned int crate, unsigned int rod, int id, int bank, int elems);

  /// Initialise a Module from the configuration.
  /**  
       Load module configuration from database
       Send module configuration to ROD

       Return 0 for success
  */
  int initialiseModule(std::string module);

  /*************************************************************
   *          Internal functions (shouldn't be exported,       *
   *              probably able to keep them private )         *
   *************************************************************/

  /// Send a primitive list to a slave DSP
  /** 
      Send a RodPrimList to the Master DSP which contains the RodPrimList which is passed as an argument

      @param await If true master waits for completion of slave primlist
      @param response If true master expects reponse from slave
   */
  void sendSlavePrimList(unsigned int partition, unsigned int crate, unsigned int rod,
                         boost::shared_ptr<PrimListWrapper> prim, unsigned int slave, bool await, bool response);

  /// Build a slave primitive list onto a master primitive list
  /** 
      Put a send slave list primitive into a primitive list
      Follow it by a start list primitive
   */
  void buildSlavePrimList(boost::shared_ptr<PrimListWrapper> primList, boost::shared_ptr<PrimListWrapper> prim, 
                          unsigned int slaveNumber, bool await, bool response);

  /// Send a primitive list to a ROD
  /**
     A helper function for other methods in this class,
     to be used only internally to this class as 
     it exposes the RodPrimList class
  */
  void sendPrimList(unsigned int partition, unsigned int crate, unsigned int rod,
                    boost::shared_ptr<PrimListWrapper> prim);

  /// Send a primitive list to all the RODs in a crate
  /**
     A helper function for other methods in this class,
     to be used only internally to this class as 
     it exposes the RodPrimList class
  */
  void sendPrimListAll(unsigned int partition, unsigned int crate,
                       boost::shared_ptr<PrimListWrapper> prim);

  /// Create histogram task primitive
  void startHistogramTask(const Scan &scan, const ScanEx &ex, boost::shared_ptr<PrimListWrapper> list);

  /// Setup event trapping according to scan and ex
  void setupEventTrapping(const Scan &scan, const ScanEx &ex, boost::shared_ptr<PrimListWrapper> list);

  /// Start event trapping according to scan and ex
  void startEventTrapping(const Scan &scan, const ScanEx &ex, boost::shared_ptr<PrimListWrapper> primList);

  /// Setup histogramming according to scan and ex
  void setupHistogramming(const Scan &scan, const ScanEx &ex, boost::shared_ptr<PrimListWrapper> list);

  /// Stop the histogramming and event trapping tasks
  void stopHistogramming(ScanEx &ex);

  /// Setup ROD for calibration
  void calib_init();

  /// Print ROD calibration registers
  void print_calib(unsigned int partition, unsigned int crate, unsigned int rod);

  /// Poll the histogramming on a ROD
  /**
     Wait for the number of bins to equal (bins - 1) on the master

     Then wait for the number of events on the slave to equal to events

     Timeout if the bin count doesn't change, and if the number 
     of events doesn't change
  */
  void pollHistogramming(ScanEx &ex, int slave, int bins, int events, int timeout = 5);

  /// Read histogram data
  scan_result_ptrs readHistogramData(const Scan &scan, const ScanEx &ex, UINT32 mid, int index);

  /// Read the histogram data in to a file
  void readHistogramsToFile(const Scan &scan, const ScanEx &ex);

  /// Write a histogram to a file
  void saveHistogramToFile(scan_result_ptrs histo, std::string filename);

#ifdef USE_IS
  /// Read the histogram data in to IS
  void readHistogramsToIS(const Scan &scan, const ScanEx &ex);
#endif

  /// Read raw data from all links into one "histogram"
  void readRawData(unsigned int partition, unsigned int crate, unsigned int rod, 
                   int delay, bool setMask);

  ///  Put raw data into FIFOs ready to read out
  int setupRawData(unsigned int partition, unsigned int crate, unsigned int rod, 
		   int delay, int units, bool setMask, const Trigger *trig = 0);

  /// Print a decoded out list
  /**
     @param outlist The list as a block of words
     @param lenth Length of the block of words
     @param in Is this an input list (hence negating the method name!
   */
  void printOutList(unsigned long *outlist, unsigned long length, bool in = false, int level = 0);

  /// Utility function to get the Crate object from the map
  /** 
      Lookup the Crate object to send instructions to.
      Return 0 if not found.
   */
  Crate *getCrate(unsigned int partition, unsigned int crate) const;

  /// Lookup module configuration in the cache
  ABCDModule *lookupConfig(UINT32 mid);

  /// Put module mid into the cache (if present in the configuration)
  void cacheModuleConfig(UINT32 mid);

  /// Translate mid to pcrc of redundant connection
  /**
     XXX What does it do on failure!!!
   */
  void getrpcrc(UINT32 mid, unsigned int &rpartition, unsigned int &rcrate, unsigned int &rrod, unsigned &rchannel);

  /// I think I have to do this after ILU is running (otherwise SCTDCSCommand can't do the stuff it needs to initialise
  void initDDC();

  /*********** Internal variables *************/

  /// The configuration database
  boost::shared_ptr<SctConfiguration::Configuration> config;

  MRSStream *mrs;
#if USE_IS
  ISInfoDictionary *m_isDict;
#endif

  IPCPartition *ddcPartition;
  string *ddcCtrls[2];
  SCTDCSCommand *ddcCmd;

  /// The current debug primitive list
  boost::shared_ptr<PrimListWrapper> debugPrimList;

  /// Map to retrieve crate objects
  std::map<std::pair<unsigned int, unsigned int>, Crate* > crateMap;

  /// Cache of module configurations
  std::map<UINT32, ABCDModule> moduleMap;

  /// Log everything that is called
  Log log;

  /// What debug information is printed
  int debugLevel;

  /// For debug only, save the scan information
  Scan *lastDebugScan;
  ScanEx lastDebugScanEx;

  /// Response code from awaitResponse
  int responseCode;

  /// Current scan number
  UINT32 scanNumber;

  /// Current run number
  UINT32 runNumber;

  /* **********  Mutexes  ********** */
  boost::mutex logMutex;
};
}  // End of namespace SctApi
#endif

