#include <unistd.h>
#include <time.h>

#include <iostream>
#include <fstream>

#include "RodCrate/RodModule.h"
#include "RodCrate/BocCard.h"
#include "RodCrate/TimModule.h"              // For TimMaskFrequency
// #include "VmeInterface/VmeInterface.h"

#include "sctConf/configXMLImpl.h"

#include "../VmeInterface/DummyVmeInterface.h"
#ifndef USE_DUMMY_VME
#include "../VmeInterface/RCCVmeInterface.h"
#endif

#include "VmeDecode.h"

#include "crate.h"

#ifdef USE_THREADS
#include <boost/bind.hpp>
#endif

// For modify BOC params
#include "ABCD/ABCDscans.h"

using namespace SctPixelRod;
using namespace SctConfiguration;

namespace SctApi {
/*
The constructor. Tell this crate which partition it belongs to.
This creates a VmeInterface.
*/
Crate::Crate(unsigned int partition, unsigned int crate, 
             boost::shared_ptr<Configuration> newConf) : 
  rodMap(), tim(0), config(newConf), vme(0), 
  partition(partition), crate(crate), mrs(0), stopPolling(false)
{
  // Create VME interface
  std::cout << "Initialising VME interface ... \n";

  if(config->isDummyCrate(partition, crate)) { 
    vme = new SctPixelRod::DummyVmeInterface();

    if(mrs) {
      *mrs << "CRATE_DUMMY"  << MRS_WARNING << MRS_QUALIF("SCTAPI") << MRS_TEXT("Using DUMMY interface") << ENDM;
    }
#ifdef USE_DUMMY_VME
  } else {
    if(mrs) {
      *mrs << "CRATE_DUMMY"  << MRS_ERROR << MRS_QUALIF("SCTAPI") << MRS_TEXT("Non dummy crate requested, support not compiled!") << ENDM;
    }

#warning "This should be SctApiException"
    throw ConfigurationException("RCC support not compiled");
  }
#else
  } else {
    vme = new SctPixelRod::RCCVmeInterface();
    if(vme->getLastErrcode() == 515) {
      // Failed to open interface (probably not running on an SBC!)
      std::cerr << "Interface not opened error, check host and VME drivers. Crate initialisation failed!!\n";

      if(mrs) {
        *mrs << "CRATEINIT_FAILED" << MRS_ERROR << MRS_QUALIF("SCTAPI") << MRS_TEXT("Crate initialisation failed (VME interface)") << ENDM;
      }

#warning "This should be SctApiException"
      throw ConfigurationException("Interface not found");
    }
    //    std::cout << "Last error " << vme->getLastErrcode() << " (" << vme->getErrorMessage(vme->getLastErrcode()) << std::endl;

    if(mrs) {
      *mrs << "CRATEINIT_SUCCESS" << MRS_INFORMATION << MRS_QUALIF("SCTAPI") << MRS_TEXT("Successfully initialised VME interface") << ENDM;
    }
  }
#endif

  std::cout << " ... Done\n";

  if(config.get() == 0) {
    std::cout << "Load configuration\n";
    config.reset(new ConfigurationXMLImpl());
  }

#if USE_THREADS
  pollThread.reset(new boost::thread(boost::bind(&Crate::pollingLoop, this)));
#endif
}

/*
The destructor. Delete all the RodModules and VmeInterface
*/
Crate::~Crate() {
  // First make sure the polling is not using them
#if USE_THREADS
  stopPolling = true;
  pollThread->join();
#endif

  for(RodMap::const_iterator i=rodMap.begin();i!=rodMap.end();i++){
    std::cout << "RodModule on shutdown " << *(i->second.second) << std::endl;
    delete i->second.first;    //RodStatus
    delete i->second.second;   //RodModule
  }
  delete tim;

  delete vme;
}

/*
  Create a TimModule object using information from the Configuration.
*/
int Crate::initialiseTim() {
  std::cout << "Load configuration for tim\n";

  TimConfig timConf;
  try {
    timConf = config->getTimConfig(partition, crate);
  } catch(ConfigurationException &c) {
    tim = 0;
    return -1;
  }

  try {
    // Create TimModule and initialize it
    std::cout << "Initialise TIM\n";

    tim = new SctPixelRod::TimModule(timConf.baseAddress, timConf.mapSize, *vme);

    std::cout << "Constructed\n";

    tim->reset();
    tim->initialize();

    std::cout << "Initialised\n";

    std::cout << "Started TIM\n";
    tim->status();
  } catch (SctPixelRod::TimException &r) {
    tim = 0;

    std::cerr << "While initialising TIM " << partition << " " << crate << ":\n";
    std::cerr << "Got TimException: " << std::endl;
    std::cerr << r.getDescriptor() << ", " << r.getData1() << ", " << r.getData2() << std::endl;
    std::cerr << "Continuing...\n";
  } catch (VmeException &v) {
    tim = 0;

    std::cout << "VmeException initialising TIM:\n";
    Utility::decodeVme(v);

    return -9;
  } catch (...) {
    tim = 0;

    std::cerr << "Unknown exception initialising TIM\n";

    return -6;
  }

  return 0;
}

/*
  Create a RodModule object using information from the Configuration.
*/
int Crate::initialiseRod(unsigned int rod) {
  int slave = 0;

  std::cout << "Load configuration for rod\n";

  if(config->isDummyCrate(partition, crate)) {
    // Don't create a RODModule!
    return 0;
  }

  SctConfiguration::RodConfig rodConf;
  try {
    rodConf = config->getRodConfig(partition, crate, rod);
  } catch(ConfigurationException &c) {
    return -1;
  }

  SctPixelRod::RodModule* rodObject;

  try {
    // Create RodModule and initialize it
    std::cout << "Initialise ROD " << rod << " at 0x" << std::hex << rodConf.baseAddress << std::dec << std::endl;

    rodObject = new SctPixelRod::RodModule(rodConf.baseAddress, 
                                           rodConf.mapSize, 
                                           *vme, 
                                           rodConf.numSlaves);

    std::cout << "Constructed\n";

    if(rodConf.resetLevel > 1) {
      std::cout << "*** ROD reset level " << rodConf.resetLevel << " (not reset) **** \n";
      std::cout << "*** Debug, ROD not reset **** \n";
      rodObject->initialize(false);  // Don't reset the ROD!
    } else {
      std::cout << "Doing full ROD reset...\n";
      // The true is a new feature to reset the ROD (what it always used to do!)
      rodObject->initialize(true);   // Use true to reset the ROD!
      std::cout << " ... done\n";
    }

    std::cout << "Initialised\n";

    std::cout << "Started ROD\n";

    rodObject->status();

    while(rodObject->textHandler() == TEXT_RQ_SET) {
      doTextBuffer(*rodObject);
    }

    int startedSlaves = 0;

    RodStatus *currStatus = new RodStatus;
    currStatus->slaves = 0;
#if USE_THREADS
    currStatus->currState = MYPRIM_IDLE;
    //     std::cout << "STATE: Initialising state to IDLE\n";
#endif

    // Initialize slaves requested by config
    try {
      for(slave=0;  slave < rodConf.numSlaves; slave++) {
        if(rodConf.slaves[slave].ipramFile != "") {
          std::cerr << "Starting slave " << slave << std::endl;
          std::cerr << "((" << rodConf.slaves[slave].emifFile << "), " 
                    << rodConf.slaves[slave].ipramFile << ", " 
                    << rodConf.slaves[slave].idramFile << ", "
                    << rodConf.slaves[slave].extFile << ")\n";

          rodObject->initSlaveDsp(rodConf.slaves[slave].ipramFile, 
                                  rodConf.slaves[slave].idramFile, 
                                  rodConf.slaves[slave].extFile, slave, 'v');

          RodOutList *outList = rodObject->getOutList();
          if(outList) {
            unsigned long* outBody = outList->getBody();

            if(outBody) {
              if(outBody[3] != PRIM_LIST_REVISION) {
                std::cout << "Prim list revision disagrees\n";
              }

              if(outBody[6] != START_SLAVE_EXECUTING) {
                std::cout << "Prim ID not start slave!\n";
              }

              if(outBody[8] != slave) {
                std::cout << "Start slave says slave 0x" << std::hex << outBody[8] << std::dec << " started not " << slave << std::endl;
              }
            } else {
              std::cout << "No body from start slave outlist!\n";
            }
          } else {
            std::cout << "No response from start slave!\n";
          }
          startedSlaves ++;
          currStatus->slaves |= 1<<slave;
        }
      }
    } catch (SctPixelRod::NoImageFile &f) {
      std::cerr << "While initialising ROD " << partition << " " << crate << " " << rod << " slave " << slave << ":\n";
      std::cerr << "No image file exception " << f.getFileName() << std::endl;
      std::cerr << "Continuing...\n";

      if(startedSlaves < 1) {
        *mrs << "ROD Slave Initialisation Failure" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
             << MRS_PARAM<const char *>("Filename", f.getFileName().c_str())
             << ENDM;
      }
    } catch (SctPixelRod::RodException &r) {
      std::cerr << "While initialising ROD " << partition << " " << crate << " " << rod << " slave " << slave << ":\n";
      std::cerr << "Bad ROD exception: " << std::endl;
      std::cerr << r.getDescriptor() << ", " << r.getData1() << ", " << r.getData2() << std::endl;
      std::cerr << "Continuing...\n";
    }

    std::cout << "Started " << startedSlaves << " ROD slave DSPs\n";

    rodMap.insert(std::make_pair(rod, std::make_pair(currStatus, rodObject)));

    std::cout << "ROD " << rod << " initialised\n";

    if(mrs) {
      *mrs << "ROD_INIT" << MRS_INFORMATION << MRS_QUALIF("SCTAPI") 
           << MRS_PARAM<int>("number", rod)
           << MRS_PARAM<int>("slaves", startedSlaves)
           << MRS_PARAM<int>("address", rodConf.baseAddress)
           << MRS_TEXT("ROD Initialised")
           << ENDM;
    }
  } catch (SctPixelRod::PrimListException &p) {
    std::cerr << "While initialising ROD " << partition << " " << crate << " " << rod << ":\n";
    std::cerr << "Exception sending primlist:\n";
    std::cerr << p.getDescriptor() << " " << p.getData1() << ", " << p.getData2() << std::endl;
    return -2;
  } catch (SctPixelRod::HpiException &h) {    
    std::cerr << "While initialising ROD " << partition << " " << crate << " " << rod << ":\n";
    std::cerr << "HpiException:\n";
    std::hex(std::cerr);
    std::cerr << h.getDescriptor() << '\n';
    std::cerr << "calcAddr: " << h.getCalcAddr() << ", readAddr: " << 
      h.getReadAddr() << '\n';
    std::dec(std::cerr);
    return -3;
  } catch (SctPixelRod::RodException &r) {
    std::cerr << "While initialising ROD " << partition << " " << crate << " " << rod << ":\n";
    std::cerr << "Rod Exception\n";
    std::cerr << r.getDescriptor() <<", " << r.getData1() << ", " << r.getData2()
              << '\n';
    return -4;
  } catch (SctPixelRod::NoImageFile &f) { // This will be caught in the slave section!
    std::cerr << "While initialising ROD " << partition << " " << crate << " " << rod << " slave " << slave << ":\n";
    std::cerr << "No image file exception " << f.getFileName() << std::endl;

    if(mrs) {
      *mrs << "RODINIT_FAIL" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_TEXT("ROD init failure")
           << MRS_PARAM<const char *>("reason" , "NoImageFile exception")
           << MRS_PARAM<int>("number", rod)
           << MRS_PARAM<int>("address", rodConf.baseAddress)
           << ENDM;
    }

    return -5;
  } catch (VmeException &v) {
    std::cout << "VmeException initialising ROD:\n";
    Utility::decodeVme(v);

    if(mrs) {
      *mrs << "RODINIT_FAIL" << MRS_ERROR << MRS_QUALIF("SCTAPI") 
           << MRS_TEXT("ROD init failure")
           << MRS_PARAM<const char *>("reason", "VmeException exception")
           << MRS_PARAM<int>("number", rod)
           << MRS_PARAM<int>("address", rodConf.baseAddress)
           << ENDM;
    }

    return -9;
  } catch (...) {
    std::cerr << "Unknown exception initialising ROD\n";
    return -6;
  }

  return 0;
}

/*
  Initialise a BOC using information from the Configuration.
*/
int Crate::initialiseBOC(unsigned int rod) {
  BocCard* bocCard;

  try {
    SctConfiguration::RodConfig rodConf = config->getRodConfig(partition, crate, rod);

    //     if(!rodConf.bocPresent) {
    //       std::cout << "Skipping BOC (configuration says not present)\n";
    //       return -2;
    //     }
  } catch(ConfigurationException &c) {
    std::cout << "Request to initialise BOC with unconfigured rod\n";
    return -1;
  }

  if(RODPresent(rod)) {
    try {
      std::cout << "Initialise BOC " << rod << std::endl;

      bocCard = getRod(rod)->getBocCard();
      bocCard->initialize();
      bocCard->status(); // temporary

    } catch (VmeException &v) {
      std::cout << "VmeException initialising BOC:\n";
      Utility::decodeVme(v);
      return -9;
    } catch (...) {
      std::cerr << "Unknown exception initialising BOC\n";
      return -6;
    }

    configureBOC(rod);
  } else {
    std::cout << "Request for non-existent BOC " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }

  return 0;
}

/*
  Set up a BOC parameters according to information from the Configuration.
*/
int Crate::configureBOC(unsigned int rod) {
  BocCard *bocCard = getRod(rod)->getBocCard();

  BOCGlobalConfig globalConf = config->getBOCGlobalConfig(partition, crate, rod);
  for(int r=0; r<9; r++) {
    if(globalConf.validMask & 1<<r) {
      switch(r) {
      case 0:
        bocCard->setClockControl(globalConf.clockControl);
        break;
      case 1:
        bocCard->setRxDataMode(globalConf.rxDataMode);
        break;
      case 2:
        if(globalConf.rxDacClear)
          bocCard->clearRxDac();
        break;
      case 3:
        if(globalConf.txDacClear)
          bocCard->clearTxDac();
        break;
      case 4:
        bocCard->setVernierFinePhase(globalConf.vernierFinePhase);
        break;
      case 5:
        bocCard->setVernierClockPhase0(globalConf.vernierClockPhase0);
        break;
      case 6:
        bocCard->setVernierClockPhase1(globalConf.vernierClockPhase1);
        break;
      case 7:
        bocCard->setBpmClockPhase(globalConf.bpmClockPhase);
        break;
      case 8:
        bocCard->setBregClockPhase(globalConf.bregClockPhase);
        break;
      }
    }
  }

  char *mappings = config->getFibreMappings(partition, crate, rod);

  for(int channel=0;channel<48;channel++){
    try {
      BOCChannelConfig channelConfig = config->getBOCConfig(partition, crate, rod, channel);
      std::cout << " Data for channel " << channel << std::endl;

      // NB - fibre mapping hopefully now not hard wired
      int txChannel = mappings[channel * 3 + 0];
      int rx0Channel = mappings[channel * 3 + 1];
      int rx1Channel = mappings[channel * 3 + 2];

      //        std::cout << txChannel << " " << rx0Channel << " " << rx1Channel << std::endl;

      // Configure clock/command output
      bocCard->setLaserCurrent(txChannel,   channelConfig.current);
      bocCard->setBpmFineDelay(txChannel,   (channelConfig.delay&0xff) );
      bocCard->setBpmCoarseDelay(txChannel, ((channelConfig.delay>>8)&0xff));
      bocCard->setBpmMarkSpace(txChannel,   channelConfig.markSpace);

      // Configure data reception
      bocCard->setRxThreshold(rx0Channel, channelConfig.threshold0);
      bocCard->setRxDataDelay(rx0Channel, channelConfig.delay0);
      bocCard->setRxThreshold(rx1Channel, channelConfig.threshold1);
      bocCard->setRxDataDelay(rx1Channel, channelConfig.delay1);
    } catch(ConfigurationException &c) {
      // This would be worth saying if a module was attached
      //         std::cout << "No channel data for BOC " << rod << " channel " << channel << std::endl;
    }
  }

  delete [] mappings;

  return 0;
}

/*!
  Send a primitive list to a ROD within this crate
*/
int Crate::sendPrimList(unsigned int rod, shared_ptr<PrimListWrapper> prim) {
  if(RODPresent(rod)) {   // If there's a rod presumably there's a data structure with it...
#if USE_THREADS
    {
      getRod(rod)->deleteOutList();
      RodStatus &stat = getRodData(rod);
      boost::mutex::scoped_lock lock(stat.mutex);
      stat.listQueue.push_back(prim);
      //       std::cout << "STATE: Changing state to LOADED\n";
      if(stat.currState == MYPRIM_IDLE) {
        stat.currState = MYPRIM_LOADED;
      } else {
        //         std::cerr << "STATE: Sending prim list when not idle!\n";
        stat.currState = MYPRIM_LOADED;
        return -10;
      }
    }  // Release lock
#else
    try {
      // So the result can't be confused with any previous
      getRod(rod)->deleteOutList();
      getRod(rod)->sendPrimList(prim->list.get());
    } catch (SctPixelRod::PrimListException &p) {
      std::cerr << "While sending primlist " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "Exception sending primlist:\n";
      std::cerr << p.getDescriptor() << " " << p.getData1() << ", " << p.getData2() << std::endl;
      return -6;
    } catch (SctPixelRod::HpiException &h) {    
      std::cerr << "While sending primlist " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "HpiException:\n";
      std::hex(std::cerr);
      std::cerr << h.getDescriptor() << '\n';
      std::cerr << "calcAddr: " << h.getCalcAddr() << ", readAddr: " << 
        h.getReadAddr() << '\n';
      std::dec(std::cerr);
      return -2;
    } catch (SctPixelRod::RodException &r) {
      std::cerr << "While sending primlist " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "Rod Exception\n";
      std::cerr << r.getDescriptor() <<", " << r.getData1() << ", " << r.getData2()
                << '\n';
      return -3;
    } catch (SctPixelRod::NoImageFile &f) {
      std::cerr << "While sending primlist " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "No image file exception " << f.getFileName() << std::endl;
      return -4;
    } catch (...) {
      std::cerr << "Unknown exception sending primlist to ROD\n";
      return -5;
    }
    //    std::cout << "Status after sendPrimList is: " << *(getRod(rod));

#endif
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }

  return 0;
}

/*!
  Send a primitive list to a ROD within this crate
*/
int Crate::sendPrimListAll(shared_ptr<PrimListWrapper> prim) {
#if USE_THREADS
  listQueueAll.push_back(prim);

  return 0;
#else
  int result = 0;
  for(RodMap::const_iterator item = rodMap.begin();
      item != rodMap.end();
      item++) {
    int rod = item->first;

    int newResult = sendPrimList(rod, prim);
    if(result == 0) result = newResult;
  }

  return result;
#endif
}

void Crate::primListControl(unsigned int index, PLControl function) {
  if(RODPresent(index)) {
    SctPixelRod::RodModule &rod = *(getRod(index));

    switch(function) {
    case C_PL_STOP:
      rod.setVmeCommandRegBit(CR_ABORT);
      break;
    case C_PL_PAUSE:
      rod.setVmeCommandRegBit(CR_PAUSE);
      break;
    case C_PL_RESUME:
      rod.setVmeCommandRegBit(CR_RESUME);
      break;
    case C_PL_CANCEL:
      rod.clearVmeCommandRegBit(CR_IN_LIST_RDY);
      break;
    }
  }
}

/*
  Await response from a ROD within this crate
*/
int Crate::awaitResponse(unsigned int rod, int timeout) {
  // Preliminary implementation from EchoTest
  // Wait for ROD to begin executing and then wait for it to finish executing
  // Check for error messages in text buffer and read them out if they exist.

  using namespace SctPixelRod;

  bool timedout = false;

  time_t start_time = time(0);

  if(RODPresent(rod)) {
#if USE_THREADS
    //     RodPrimState returnPState;
    /* Using threads we let the polling thread handle the text buffers
       and all the communication with the ROD 
    */

    //     SctPixelRod::RodModule &myRod = *getRod(rod);
    RodStatus &status = getRodData(rod);

    try {
      RodPrimState localState;
      do {
        if((time(0) - start_time) > timeout) {
          timedout = true;
          std::cout << "Timeout from wait complete\n";
        } else {
        }

        // Yield when haven't got the lock...
        sched_yield();

        {
          boost::mutex::scoped_lock statLock(status.mutex);
          localState = status.currState;
        }
      } while ((localState != MYPRIM_COMPLETE)
               &&(!timedout));
      //      std::cout << "AwaitResponse: done wait or idle " << status.currState << "\n";

      {
        boost::mutex::scoped_lock statLock(status.mutex);
        status.currState = MYPRIM_IDLE;
        //         std::cout << "STATE: Changing state to IDLE\n";

        //         std::cout << "Current length of queue " << status.listQueue.size() << std::endl;
        //         std::cout << "Current RodModule state  " << getRod(rod)->getPrimState() << std::endl;
      }

      if(timedout) std::cout << "Timed out at " << timeout << " seconds\n";
    } catch (BaseException &b) {
      std::cerr << "In await response " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "Exception: " << b << std::endl;
      return -6;
    } catch (...) {
      std::cerr << "Unknown exception sending primlist to ROD\n";
      return -2;
    }
#else
    // Create a buffer for test messages
    PrimState returnPState;
    TextBuffState returnTState;

    SctPixelRod::RodModule* rod0 = getRod(rod);

    try {
      do {
        returnPState = rod0->primHandler();
        returnTState = rod0->textHandler();
        while (returnTState == TEXT_RQ_SET) {
          doTextBuffer(*rod0);

          // Check there isn't another text buffer
          returnTState = rod0->textHandler();
        }

        if((time(0) - start_time) > timeout) {
          timedout = true;
          std::cout << "Timeout from wait execute\n";
        }
      } while ((returnPState != PRIM_EXECUTING) && (!timedout));
      do {
        returnPState = rod0->primHandler();

        if((time(0) - start_time) > timeout) {
          timedout = true;
          std::cout << "Timeout from wait finish\n";
        }
      } while ((returnPState != PRIM_WAITING)&&(returnPState != PRIM_IDLE) &&(!timedout));

      if(timedout) std::cout << "Timed out at " << timeout << " seconds\n";
    } catch (PrimListException &p) {
      std::cerr << "In await response " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "Exception sending primlist:\n";
      std::cerr << p.getDescriptor() << " " << p.getData1() << ", " << p.getData2() << std::endl;
      return -6;
    } catch (HpiException &h) {    
      std::cerr << "In await response " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "HpiException:\n";
      std::hex(std::cerr);
      std::cerr << h.getDescriptor() << '\n';
      std::cerr << "calcAddr: " << h.getCalcAddr() << ", readAddr: " << 
        h.getReadAddr() << '\n';
      std::dec(std::cerr);
      return -5;
    } catch (VmeException &v) {
      std::cout << "VmeException in await response:\n";
      Utility::decodeVme(v);

      return -9;
    } catch (RodException &r) {
      std::cerr << "In await response " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "Rod Exception\n";
      std::cerr << r.getDescriptor() <<", " << r.getData1() << ", " << r.getData2()
                << '\n';
      return -4;
    } catch (NoImageFile &f) {
      std::cerr << "In await response " << partition << " " << crate << " " << rod << ":\n";
      std::cerr << "No image file exception " << f.getFileName() << std::endl;
      return -3;
    } catch (...) {
      std::cerr << "Unknown exception sending primlist to ROD\n";
      return -2;
    }
    //    std::cout << "Status at end of awaitResponse: " << *rod0;
#endif
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;

    return -1;
  }

  if(timedout)
    return -200;

  return 0;
}

/*
  Await response from all RODs within this crate
*/
int Crate::awaitResponseAll(int timeout) {
  int result = 0;
  for(RodMap::const_iterator item = rodMap.begin();
      item != rodMap.end();
      item++) {
    int rod = item->first;

    int newResult = awaitResponse(rod, timeout);
    if(result == 0) result = newResult;
  }

  return 0;
}

/*
  Get the last response from a ROD within this crate
*/
shared_ptr<RodOutList> Crate::getResponse(unsigned int rod) {
  RodOutList *result = 0;

  if(RODPresent(rod)) {
    RodOutList *response = getRod(rod)->getOutList();

    if(response) {
      result = new RodOutList(response->getLength());

      unsigned long *oldBody = response->getBody();
      unsigned long *newBody = result->getBody();

      for(int i=0; i<result->getLength(); i++) {
        newBody[i] = oldBody[i];
      }
    } else {
      std::cout << "**** No response found when expected\n";
      result = 0;
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    result = 0;
  }

  return shared_ptr<RodOutList>(result);
}

void Crate::setMrsStream(MRSStream *stream) 
{
  mrs = stream;
}

/*
  Read the message buffer if there is anything there
*/
bool Crate::getRodMessage(unsigned int rod, char *buffer, int &length) {
  bool complete = true;

  char myBuffer[TEXT_BUFF_SIZE];
  long myLength;

  TEXT_BUFFER_TYPE dummy;
  if(RODPresent(rod)) {
    if(getRod(rod)->textHandler() == TEXT_READOUT) {
      getRod(rod)->getTextBuffer(myBuffer, myLength, dummy);
      getRod(rod)->clearTextBuffer();

      // Copy across what's allowed
      for(int i=0; i<length || i<myLength; i++) {
        buffer[i] = myBuffer[i];
      }

      if(length < myLength) {
        complete = false;
      } else {
        length = myLength;
      }
    } else {
      buffer[0] = '\0';
      length = 0;
      complete = false;
    }
  }

  return complete;
}

/*
  Write a block of master dsp memory 
*/
int Crate::mdspBlockWrite(unsigned int rod, long dspStart, unsigned long* buffer, long numWords) {
  if(RODPresent(rod)) {
    try {
      getRod(rod)->mdspBlockWrite(dspStart, buffer, numWords);
    } catch(BaseException &b) {
      std::cout << "Exception writing mdsp\n";
      std::cout << b;
      return -3;
    } catch(...) {
      std::cout << "Exception writing mdsp\n";
      return -2;
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }

  return 0;
}

/*
  Read a block of master dsp memory 
*/
int Crate::mdspBlockRead(unsigned int rod, long dspStart, unsigned long* buffer, long numWords) {
  if(RODPresent(rod)) {
    try {
      std::cout << "Calling RodModule::mdspBlockRead(0x" << std::hex << dspStart << ", " << buffer << ", " << std::dec << numWords << ");\n";
      getRod(rod)->mdspBlockRead(dspStart, buffer, numWords);
    } catch(BaseException &b) {
      std::cout << "Exception reading mdsp\n";
      std::cout << b;
      return -3;
    } catch(...) {
      std::cout << "Exception reading mdsp\n";
      return -2;
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }

  return 0;
}

/*
  Read a block of slave dsp memory 
*/
int Crate::slvBlockWrite(unsigned int rod, long dspStart, unsigned long* buffer, long numWords, long dspNumber) {
  if(RODPresent(rod)) {
    try {
      getRod(rod)->slvBlockWrite(dspStart, buffer, numWords, dspNumber);
    } catch(BaseException &b) {
      std::cout << "Exception writing to slave\n";
      std::cout << b;
      return -3;
    } catch(...) {
      std::cout << "Exception writing to slave\n";
      return -2;
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }
  return -0;
}

/*
  Read a block of slave dsp memory 
*/
int Crate::slvBlockRead(unsigned int rod, long dspStart, unsigned long* buffer, long numWords, long dspNumber) {
  if(RODPresent(rod)) {
    try {
      getRod(rod)->slvBlockRead(dspStart, buffer, numWords, dspNumber);
    } catch(BaseException &b) {
      std::cout << "Exception reading slave\n";
      std::cout << b;
      return -3;
    } catch(...) {
      std::cout << "Exception reading slave\n";
      return -2;
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }
  return 0;
}

unsigned long Crate::readRodStatusReg(unsigned int rod, long regNumber) {
  unsigned long result = 0;

  if(RODPresent(rod)) {
    try {
      result = getRod(rod)->readRodStatusReg(regNumber);
    } catch(BaseException &b) {
      std::cout << "Exception reading rod status\n";
      std::cout << b;
    } catch(...) {
      std::cout << "Exception reading rod status\n";
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
  }
  return result;
}

unsigned long Crate::readRodCommandReg(unsigned int rod, long regNumber) {
  unsigned long result = 0;

  if(RODPresent(rod)) {
    try {
      result = getRod(rod)->readRodCommandReg(regNumber);
    } catch(BaseException &b) {
      std::cout << "Exception reading rod command\n";
      std::cout << b;
    } catch(...) {
      std::cout << "Exception reading rod command\n";
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
  }
  return result;
}

unsigned long Crate::dspSingleRead(unsigned int rod, const unsigned long dspAddr, long dspNumber) {
  unsigned long result = 0;

  if(RODPresent(rod)) {
    try {
      if(dspNumber == -1) {
        result = getRod(rod)->mdspSingleRead(dspAddr);
      } else {
        result = getRod(rod)->slvSingleRead(dspAddr, dspNumber);
      } 
    } catch(BaseException &b) {
      std::cout << "Exception dsp single read\n";
      std::cout << b;
    } catch(...) {
      std::cout << "Exception dsp single read\n";
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
  }
  return result;
}

int Crate::dspSingleWrite(unsigned int rod, unsigned long dspAddr, unsigned long buffer, long dspNumber) {
  if(RODPresent(rod)) {
    try {
      if(dspNumber == -1) {
        getRod(rod)->mdspSingleWrite(dspAddr, buffer);
      } else {
        getRod(rod)->slvSingleWrite(dspAddr, buffer, dspNumber);
      }
    } catch(BaseException &b) {
      std::cout << "Exception dsp single write\n";
      std::cout << b;
    } catch(...) {
      std::cout << "Exception dsp single write\n";
      return -2;
    }
  } else {
    std::cout << "Request for non-existent ROD " << partition << " " << crate << " " << rod << std::endl;
    return -1;
  }

  return 0;
}

void Crate::status() {
  for(RodMap::const_iterator item = rodMap.begin();
      item != rodMap.end();
      item++) {
    RodModule &rod = *(item->second.second);

    try {
      SctConfiguration::RodConfig rodConf = config->getRodConfig(partition, crate, item->first);

      //       if(rodConf.bocPresent) {
      std::cout << "BOC status\n";
      rod.getBocCard()->status();
      //       } else {
      //         std::cout << "*********** No BOC status (BOC not present!!!!\n";
      //       }
    } catch(ConfigurationException &c) {
      std::cout << "Request for status on unconfigured ROD! (impossible)\n";
      return;
    }

    std::cout << "ROD status\n";
    rod.status();

    if(rod.readRodStatusBit(0, SR_RUNNING)) {
      std::cout << "Running ";
    }
    if(rod.readRodStatusBit(0, SR_BUSY)) {
      std::cout << "Busy ";
    }
    if(rod.readRodStatusBit(0, SR_PAUSED)) {
      std::cout << "Paused ";
    }
    if(rod.readRodStatusBit(0, SR_OUT_LIST_RDY)) {
      std::cout << "Out list ready ";
    }
    if(rod.readRodStatusBit(0, SR_EXECUTING)) {
      std::cout << "Executing \n";
      unsigned long sr2 = rod.readRodStatusReg(1);

      std::cout << "Current Primitive index = " << ((sr2 & 0xfffff0) >> 4) << std::endl;
      std::cout << "Current List index = " << (sr2 & 0xf);
    }

    std::cout << std::endl;
  }
}

/* 
   Lookup the RodModule object to send instructions to.
   Return 0 if not found.
*/
SctPixelRod::RodModule *Crate::getRod(unsigned int rod) const {
  RodMap::const_iterator found;

  found = rodMap.find(rod);
  if(found == rodMap.end()) return 0;
  else return found->second.second;
}

const RodStatus &Crate::getRodData(unsigned int rod) const {
  RodMap::const_iterator found;

  found = rodMap.find(rod);
  ///@todo Proper exception here!
  if(found == rodMap.end()) {
    std::cerr << "AJB ERROR Hello world" 
              <<  __FILE__ << ":" <<__LINE__ << std::endl;
  }
  else return *(found->second.first);
}

RodStatus &Crate::getRodData(unsigned int rod) {
  RodMap::const_iterator found;

  found = rodMap.find(rod);
  ///@todo Proper exception here!
  if(found == rodMap.end()) {
    std::cerr << "AJB ERROR Hello world" 
              <<  __FILE__ << ":" <<__LINE__ << std::endl;
  }
  else return *(found->second.first);
}

void Crate::doTextBuffer(SctPixelRod::RodModule &rod) {
  TEXT_BUFFER_TYPE myTextType;
  TextBuffState returnTState;

  std::cout << "Got some text\n";
  do {
    returnTState = rod.textHandler();
  } while (returnTState != TEXT_READOUT);

  if(mrs) {
    *mrs << "TEXTBUFF_ARRIVE" << MRS_DIAGNOSTIC << MRS_QUALIF("SCTAPI");
  }

  char * myTextBuffer = new char[TEXT_BUFF_SIZE];
  long myTextLength;

  rod.getTextBuffer(myTextBuffer, myTextLength, myTextType);
  rod.clearTextBuffer();

  std::string type;
  switch(myTextType) {
  case TEXT_ERR: type = "ERROR"; break;
  case TEXT_INFO: type = "INFO"; break;
  case TEXT_DIAG: type = "DIAG"; break;
  case TEXT_XFER: type = "TRANSFER"; break;
  case TEXT_UNDEF: type = "**Undefined**"; break;
  }

  std::cout << "Text " << type << " : " << myTextLength << std::endl;
  if(mrs) {
    *mrs << MRS_PARAM<const char *>("Type", type.c_str());
  }

  for (int i=0; i<myTextLength; i++) {
    std::cout << myTextBuffer[i];
  }

  myTextBuffer[myTextLength] = 0;
  if(mrs) {
    *mrs << MRS_TEXT(myTextBuffer);
  }

  std::cout << std::endl;

#warning "This should use debug level instead"
#if DEBUG
  std::ofstream outFile((type + "BufferDump.txt").c_str(), std::ios::binary | std::ios::app);
  outFile << "\nBuffer block length " << myTextLength << std::endl;
  time_t currTime = time(0);
  outFile << "Time " << ctime(&currTime);
  outFile.write(myTextBuffer, myTextLength);
#endif

  if(mrs) {
    *mrs << ENDM;
  }

  delete [] myTextBuffer;
}

void Crate::printBOCSetup(unsigned int rod) {
  if(RODPresent(rod)) {
    BocCard &bocCard = *(getRod(rod)->getBocCard());

    std::cout.fill('0');
    for(int c=0; c<48; c++) {
      std::cout << "Channel ";
      std::cout.width(2);
      std::cout << c << ":  I 0x";
      std::cout << std::hex;
      std::cout.width(2);
      std::cout << bocCard.getLaserCurrent(c) << " Delay 0x";
      std::cout.width(2);
      std::cout << bocCard.getBpmFineDelay(c) << "/0x";
      std::cout.width(2);
      std::cout << bocCard.getBpmCoarseDelay(c)
                << " M/S 0x";
      std::cout.width(2);
      std::cout << bocCard.getBpmMarkSpace(c);
      std::cout << "  Input 0: Th 0x";
      std::cout.width(2);
      std::cout << bocCard.getRxThreshold(c*2) << " Delay 0x";
      std::cout.width(2);
      std::cout << bocCard.getRxDataDelay(c*2)
                << "  Input 1: Th 0x";
      std::cout.width(2);
      std::cout << bocCard.getRxThreshold(c*2+1) << " Delay 0x";
      std::cout.width(2);
      std::cout << bocCard.getRxDataDelay(c*2+1)
                << std::endl;
    }
    std::cout.fill(' ');
    std::cout << std::dec;
  }
}

void Crate::printBOCRegisters(unsigned int rod) {
  if(RODPresent(rod)) {
    BocCard &bocCard = *(getRod(rod)->getBocCard());

    std::cout << std::hex;

    std::cout << "BOC register dump\n";

    std::cout << "Clock control  0x" << bocCard.getClockControl() << std::endl;
    std::cout << "RX data mode   0x" << bocCard.getRxDataMode() << std::endl;
    std::cout << "RX DAC clear   0x" << bocCard.getRxDacClear() << std::endl;
    std::cout << "TX DAC clear   0x" << bocCard.getTxDacClear() << std::endl;
    std::cout << "Vernier fine   0x" << bocCard.getVernierFinePhase() << std::endl;
    std::cout << "Vernier clk 0  0x" << bocCard.getVernierClockPhase0() << std::endl;
    std::cout << "Vernier clk 1  0x" << bocCard.getVernierClockPhase1() << std::endl;
    std::cout << "BPM clk phase  0x" << bocCard.getBpmClockPhase() << std::endl;
    std::cout << "BReg clk phase 0x" << bocCard.getBregClockPhase() << std::endl;
    std::cout << "BOC reset      0x" << bocCard.getBocReset() << std::endl;
    std::cout << "BPM reset      0x" << bocCard.getBpmReset() << std::endl;
    std::cout << "BOC status     0x" << bocCard.getBocStatusRegister() << std::endl;

    std::cout << std::dec;
  } else {
    std::cout << "No ROD found\n";
  }
}

// Check that the BOC lasers are on
/*
  status = 0xe2
  status = 0x82 when crate interlock not off
  status = 0xa0 when on board interlock not off
*/
bool Crate::checkBOCLasersOn(unsigned int rod) { 
  //   unsigned long status = getRod(rod)->getBocCard()->getBocStatusRegister();

  //   return status == 0xe2;
  if(RODPresent(rod))
    return getRod(rod)->getBocCard()->getInterlockStatus();
  else
    return false; // Ie it won't work ... (but because the ROD's not configured...)
}

void Crate::enterBOCClockBy2Mode(unsigned int rod) {
  if(RODPresent(rod)) {
    BocCard &bocCard = *(getRod(rod)->getBocCard());

    bocCard.setRxDataMode(1);
    bocCard.setClockControl(2);
  }
}

void Crate::leaveBOCClockBy2Mode(unsigned int rod) {
  if(RODPresent(rod)) {
    BocCard &bocCard = *(getRod(rod)->getBocCard());

    bocCard.setRxDataMode(0);
    bocCard.setClockControl(0);
  }
}

void Crate::modifyBOCParam(unsigned int rod, unsigned int channel, unsigned int type, unsigned int val) {
  if(RODPresent(rod)) {
    BocCard &bocCard = *(getRod(rod)->getBocCard());

    char *mappings;

    try { 
      mappings = config->getFibreMappings(partition, crate, rod);
    }  catch(ConfigurationException &c) {
      std::cout << "No fibre mappings for appropriate module: \n";
      std::cout << c.what() << std::endl;

      //     log << "  (Not found)\n";

      return;
    }

    int txFibre = mappings[channel * 3];
    int rxFibre1 = mappings[channel * 3 + 1];
    int rxFibre2 = mappings[channel * 3 + 2];

    switch(type) {
    case ST_RX_DELAY0:
      bocCard.setRxDataDelay(rxFibre1, val);
      break;
    case ST_RX_DELAY1:
      bocCard.setRxDataDelay(rxFibre2, val);
      break;
    case ST_RX_DELAY:
      bocCard.setRxDataDelay(rxFibre1, val);
      bocCard.setRxDataDelay(rxFibre2, val);
      break;

    case ST_RX_THRESHOLD0:
      bocCard.setRxThreshold(rxFibre1, val);
      break;
    case ST_RX_THRESHOLD1:
      bocCard.setRxThreshold(rxFibre2, val);
      break;
    case ST_RX_THRESHOLD:
      bocCard.setRxThreshold(rxFibre1, val);
      bocCard.setRxThreshold(rxFibre2, val);
      break;

    case ST_TX_CURRENT:
      bocCard.setLaserCurrent(txFibre, val);
      break;
    case ST_TX_MARKSPACE:
      bocCard.setBpmMarkSpace(txFibre, val);
      break;
    case ST_TX_DELAY:
      bocCard.setBpmCoarseDelay(txFibre, (val&0xff00) >> 8);
      bocCard.setBpmFineDelay(txFibre, (val&0xff));
      break;
    case ST_TX_COARSE:
      bocCard.setBpmCoarseDelay(txFibre, val);
      break;
    case ST_TX_FINE:
      bocCard.setBpmFineDelay(txFibre, val);
      break;

#ifdef ST_BOC_BPM_PH
    case ST_BOC_BPM_PH:
      bocCard.setBpmClockPhase(val);
      break;
    case ST_BOC_BREG_PH:
      bocCard.setBregClockPhase(val);
      break;
    case ST_BOC_V0_PH:
      bocCard.setVernierClockPhase0(val);
      break;
    case ST_BOC_V1_PH:
      bocCard.setVernierClockPhase1(val);
      break;
    case ST_BOC_VRN_FINE:
      bocCard.setVernierFinePhase(val);
      break;
#else
#warning "No BOC global modification"
#endif
    default:
      break;
    }
  }
}

void Crate::lasersOff() {
  for(RodMap::const_iterator ri = rodMap.begin();
      ri!=rodMap.end();
      ri++) {
    BocCard &bocCard = *(ri->second.second->getBocCard());

    for(int i=0; i<48; i++) {
      bocCard.setLaserCurrent(i, 0);
    }
  }
}

void Crate::freeTriggers(TimMaskFrequency freq) {
  if(tim)
    tim->intTrigStart(freq);
}

void Crate::stopTriggers() {
  if(tim)
    tim->intTrigStop();
}

void Crate::timL1A() {
  if(tim)
    tim->issueCommand(SctPixelRod::TIM_VTRG);
}

void Crate::timCalL1A(int delay) {
  if(tim)
    tim->issueVCAL(delay);
}

bool Crate::slavePresent(int rod, int index) const {
  bool present;

  if(!RODPresent(rod)) 
    present = false;   // The ROD's not present so the slave definitely isn't!
  else {
    present = (getRodData(rod).slaves & (1<<index))?true:false;
  }

  return present;
}

bool Crate::RODPresent(int rod) const {
  return getRod(rod);
}

} // End of namespace SctApi
