#ifndef SCTAPI_CRATE
#define SCTAPI_CRATE

#include <list>
#include <map>
#include <memory>
#include <mrs/message.h>

#include <boost/thread.hpp>
#include <boost/shared_ptr.hpp>

#include "primListWrapper.h"

#include "RodCrate/TimModule.h"              // For TimMaskFrequency
#if USE_THREADS
enum RodPrimState { MYPRIM_LOADED, MYPRIM_EXECUTING, MYPRIM_COMPLETE, MYPRIM_IDLE };
#endif

using boost::shared_ptr;

namespace SctPixelRod {
  class RodOutList;
  class RodModule;
/*   class TimModule; */   // Included anyway
}

class VmeInterface;

namespace SctConfiguration {
  class Configuration;
}

namespace SctApi {
  struct RodStatus {
#if USE_THREADS
    std::list<boost::shared_ptr<PrimListWrapper> > listQueue;
    boost::mutex mutex;
    RodPrimState currState;
#endif
    int slaves;
  };

/**
   A representation of the crate controller.

   This is used by SCTAPI to send commands to RodModule
   objects. These are stored in crate memory. 

   Module configuration will eventually be stored here 
   also?
 */
class Crate {
  // Don't use copy constructor or assignment operator
  Crate(const Crate&);
  Crate &operator=(const Crate&);
 public:
  /// The constructor. 
  /**
     Tell this crate which partition it belongs to.
     This creates a VmeInterface.
  */
  Crate(unsigned int partition, unsigned int crate, 
        boost::shared_ptr<SctConfiguration::Configuration> newConf = boost::shared_ptr<SctConfiguration::Configuration>() );
  /// The destructor
  /**
     Delete all the RodModules and VmeInterface 
  */
  ~Crate();

  /************** Interaction with Configuration **************/

  /// Initialise the TIM from the configuration.
  /**
     Return 0 for success
   */ 
  int initialiseTim();

  /// Initialise a ROD from the configuration.
  /**
     Store object in local map.
     Return 0 for success
   */ 
  int initialiseRod(unsigned int rod);

  /// Initialise a BOC from the configuration.
  /**
     Return 0 for success
  */ 
  int initialiseBOC(unsigned int rod);

  /// Configure BOC parameters from the configuration
  /**
     Return 0 for success
   */
  int configureBOC(unsigned int rod);

  /*************** Primlists *********************/

  /// Send a primitive list to a ROD
  /**
     Return 0 for success
  */
  int sendPrimList(unsigned int rod, shared_ptr<PrimListWrapper> prim);

  /// Send a primitive list to all RODs in a crate
  /**
     Return 0 for success
  */
  int sendPrimListAll(shared_ptr<PrimListWrapper> prim);

  /// Await a response from a ROD
  /**
     Await response from a ROD within this crate.
     Handle text messages sent by the ROD (print to stdout)

     If waiting longer than timeout seconds then return anyway
  */
  int awaitResponse(unsigned int rod, int timeout);

  /// Await a response from all RODs in a crate
  /**
     Handle text messages sent by the RODs (print to stdout)

     If waiting longer than timeout seconds then return anyway
  */
  int awaitResponseAll(int timeout);

  /// Get the response from a ROD
  /**
     Return the response to the last primlist (a wrapped copy)
   */
  shared_ptr<SctPixelRod::RodOutList> getResponse(unsigned int rod);

  void setMrsStream(MRSStream *stream);

  /***************** Messages ******************/

  /// Get a message in the text buffer
  /**
     @param rod Index of ROD
     @param buffer Buffer to write text into
     @param length Length of buffer allocated, on return contains amount of data copied.

     Return message from ROD if present
   */
  bool getRodMessage(unsigned int rod, char *buffer, int &length);

  /************** Misc read/writes *****************/

  /// Read data from the master DSP's memory
  /**
     Proxy for RodModule method
  */
  int mdspBlockRead(unsigned int rod, long dspStart, unsigned long *buffer, long numWords);

  /// Read data from a slave DSP's memory
  /**
     Proxy for RodModule method
  */
  int slvBlockRead(unsigned int rod, long dspStart, unsigned long *buffer, long numWords, long slaveNumber);

  /// Write data direct to a slave DSP's memory
  /**
     Proxy for RodModule method
  */
  int slvBlockWrite(unsigned int rod, long dspStart, unsigned long* buffer, long numWords, long dspNumber);

  /// Write block of master dsp memory
  /**
     Proxy for RodModule method
   */
  int mdspBlockWrite(const unsigned int rod, const long dspStart, unsigned long *buffer, const long numWords);

  /// Read Status register 
  /**
     Proxy for RodModule method
  */
  unsigned long readRodStatusReg(unsigned int rod, long regNumber);


  /// Read command register
  /**
     Proxy for RodModule method
  */
  unsigned long readRodCommandReg(unsigned int rod, long regNumber);

  /// Read single word from ROD
  /**
     Proxy for RodModule method
  */
  unsigned long dspSingleRead(unsigned int rod, const unsigned long dspAddr, long dspNumber);

  /// Write single word
  /**
     Proxy for RodModule method
  */
  int dspSingleWrite(unsigned int rod, unsigned long dspAddr, unsigned long buffer, long dspNumber);

  /// Print status of rods on crate 
  void status();

  /// Primitive list control constants 
  /**
     C_PL_STOP        Abort the current primitiveList
     C_PL_PAUSE       Pause the current primitiveList
     C_PL_RESUME      Resume the current primitiveList
     C_PL_CANCEL      Cancel the recently sent primitiveList (only if not started execution)
  */
  enum PLControl {
    C_PL_STOP,
    C_PL_PAUSE,
    C_PL_RESUME,
    C_PL_CANCEL };

  /// Control the primitive list running on a rod (Untested)
  void primListControl(unsigned int rod, PLControl function);

  /// Print all BOC paramters as set up
  void printBOCSetup(unsigned int rod);

  /// Print all BOC registers
  void printBOCRegisters(unsigned int rod);

  /// Check BOC laser interlock
  bool checkBOCLasersOn(unsigned int rod);

  /// Setup BOC to sample at 20MHz
  void enterBOCClockBy2Mode(unsigned int rod);

  /// Restore BOC to 40MHz mode
  void leaveBOCClockBy2Mode(unsigned int rod);

  /// Modify parameters on BOC (type undefined)
  void modifyBOCParam(unsigned int rod, unsigned int channel, unsigned int type, unsigned int val);

  /// Turn all BOC laser currents to 0
  void lasersOff();

  /* ******************* TIM Functions ***************************/

  /// Start the TIM generating regular triggers 
  void freeTriggers(SctPixelRod::TimMaskFrequency freq = SctPixelRod::TIM_MASK_TRIG_10_0KHZ);

  /// Stop the TIM generating regular triggers
  void stopTriggers();

  /// Tell the TIM to send an L1A 
  void timL1A();

  /// Tell TIM to send Cal + L1A
  void timCalL1A(int delay);

  /// Check if a slave has been started
  bool slavePresent(int rod, int index) const;

  /// Check if a ROD has been configured
  bool RODPresent(int rod) const;

 private:
  /// Lookup RodModule object
  /**
     Return 0 if not found
  */
  SctPixelRod::RodModule *getRod(unsigned int rod) const;

  const RodStatus &getRodData(unsigned int rod) const;
  RodStatus &getRodData(unsigned int rod);

  /// Get a text buffer from a ROD
  void doTextBuffer(SctPixelRod::RodModule &rod0);

#if USE_THREADS
  void pollingLoop();
#endif

  /// Shortcut for rod map
  typedef std::map<unsigned int, std::pair<RodStatus *, SctPixelRod::RodModule *> > RodMap;

  /// Map to lookup RodModule objects
  RodMap rodMap;

  /// TIM module
  SctPixelRod::TimModule *tim;

  /// The configuration
  boost::shared_ptr<SctConfiguration::Configuration> config;

  /// The vmeinterface that is used by all rods and tims
  SctPixelRod::VmeInterface *vme;

  /// Which partition this crate is in
  const unsigned int partition;

  /// The id of this crate
  const unsigned int crate;

  MRSStream *mrs;

  // Leave this in for no threads (its nice and small...)
  bool stopPolling;

#if USE_THREADS
  /// Queue for prim lists to go to all RODs
  std::list<shared_ptr<PrimListWrapper> > listQueueAll;

  /* **********  Mutexes  ********** */
  boost::mutex vmeMutex;

  /* ******* Threads ******/
  std::auto_ptr<boost::thread> pollThread;
#endif
};

}  // End of namespace SctApi
#endif
