#include "Sct/IS/IOManagerIS.h"
#include "Sct/Exception.h"
#include "Sct/LogicErrors.h"
#include "Sct/SctNames.h"
#include "Sct/SctParameters.h"
#include "SctData/NPtGainTestResult.h"
#include "SctData/ModuleDefectList.h"
#include "SctData/ModuleElement.h"
#include <is/isinfo.h>
#include <boost/shared_ptr.hpp>
#include <iostream>
#include <fstream>

#include "Comp.h"

using namespace Sct;
using namespace Sct::IS;
using namespace boost;
using namespace SctData;
using namespace std;

string testName;
string compFile;
double gainTol = 1;
double offsetTol = 1;
double noiseTol = 34;
bool debug = false;
Comp cGain(gainTol);
Comp cOffset(offsetTol);
Comp cNoise(noiseTol);
int noiseFailCount = 0;
int offsetFailCount = 0;
int gainFailCount = 0;
int codeFailCount = 0;
int nFail = 75;
    
void waitFor(string name, int nsecs) {
    ISInfoDictionary d(SctNames::getPartition());
    int tries = 0;
    while (!d.contains(name.c_str()) && tries < nsecs) {
	if (!sleep(1))
	    ++tries;
    }
    if (tries == nsecs) {
	throw IllegalStateError("Didn't wait long enough", __FILE__, __LINE__);
    }
    if (debug) {
	cout << "Waited for " << tries << " secs" << endl;
    }
}

void compareChannel(unsigned int ichannel, const ModuleDefectList& defects, const NPtGainTestResultData& data, istream& file) {
    unsigned int channel = 0;
    unsigned int code = 0;
    double gain = 0;
    double offset = 0;
    double noise = 0;
    string comment;
    file >> channel >> code >> gain >> offset >> noise >> comment;
    //Ignore rest of line
    file.ignore(256, '\n');
 
    if (channel != ichannel) {
	ostringstream oss;
	oss << "Uh oh - channel read is not same as channel given.  Read: " << channel << " given: " << ichannel;
	throw IllegalStateError(oss.str(), __FILE__, __LINE__);
    }
    
    //Check defects
    if (code == 0) {
	if (defects.defectAffectingElement(Channel(channel))) {
	    ++codeFailCount;
	    cout << "SCTDAQ had no defects but we found some for channel " << channel << ".  They are: " << endl;
	    defects.getDefectsAffectingElement(Channel(channel))->print(cout);
	} else if (debug) {
	    cout << "Both SCTDAQ and ourselves have no defects!" << endl;
	}
    } else {
	if (!defects.defectAffectingElement(Channel(channel))) {
	    ++codeFailCount;
	    cout << "We didn't find defects that SCTDAQ did for channel " << channel << ".  SCTDAQ found: " << endl;
	    cout << "Code: " << code  << "  " << comment << endl;
	} else {
	    cout << "Both found defects for channel " << channel << ".  SCTDAQ found: " << endl;
	    cout << "Code: " << code  << "  " << comment << ".  We found: " << endl;
	    defects.getDefectsAffectingElement(Channel(channel))->print(cout);
	}
    }
    
    //Check values
    if (cNoise.comp(data.noise, noise)) {
	++noiseFailCount;
	cout << "Noise different for channel " << channel << " SCTDAQ: " << noise << " Us: " << data.noise << endl;
    } else if (debug) {
	cout << "Noise same for channel " << channel << " SCTDAQ: " << noise << " Us: " << data.noise << endl;
    }
    
    if (cGain.comp(data.gain, gain)) {
	++gainFailCount;
	cout << "Gain different for channel " << channel << " SCTDAQ: " << gain << " Us: " << data.gain << endl;
    } else if (debug) {
	cout << "Gain same for channel " << channel << " SCTDAQ: " << gain << " Us: " << data.gain << endl;
    }
    
    if (cOffset.comp(data.offset, offset)) {
	++offsetFailCount;
	cout << "Offset different for channel " << channel << " SCTDAQ: " << offset << " Us: " << data.offset << endl;
    } else if (debug) {
	cout << "Offset same for channel " << channel << " SCTDAQ: " << offset << " Us: " << data.offset << endl;
    }
}

bool compare(shared_ptr<NPtGainTestResult> result) {    
    //Open file and loop over channels
    ifstream file (compFile.c_str());
    if (!file.good()) throw IllegalStateError("Failed to open comparison file: " + compFile, __FILE__, __LINE__);
    //Dump first line
    file.ignore(256, '\n');
    
    for (unsigned int i=0; i<nChannelModule; ++i) {
	compareChannel(i, result->getDefects(), result->getChannelData(i), file);
    }
    
    cout << "Summary: " << endl << "Gain av diff " << cGain.mean() << " +- " << cGain.stddev() <<endl;
    cout << "Noise av diff " << cNoise.mean() << " +- " << cNoise.stddev() <<endl;
    cout << "Offset av diff " << cOffset.mean() << " +- " << cOffset.stddev() <<endl;
    cout << noiseFailCount << " channels failed noise comparison" << endl;
    cout << gainFailCount << " channels failed gain comparison" << endl;
    cout << offsetFailCount << " channels failed offset comparison" << endl;
    cout << codeFailCount << " channels failed defect comparison" << endl;
    
    return (noiseFailCount + gainFailCount + offsetFailCount + codeFailCount) < nFail;
}

void help() {
    cout << "Takes 2 arguments." << endl << "The first is the name of the Test object in IS" << endl
	<< "The second is the filename of the data to compare with (SCT DB format)" << endl
        << "Takes an optional 3rd argument which indicates that it is comparing 3pt gain data" << endl;
    exit (1);
}

int main(int argc, char** argv) {
    Sct::setExceptionHandlers(argv[0]);
    if (argc < 3) help();
    testName = argv[1];
    compFile = argv[2];
    if (argc > 3) {
	cOffset.setTolerance(2);
	cNoise.setTolerance(42);
    }
    
    try {
	waitFor(testName, 60);
	shared_ptr<Serializable> ob = IOManagerIS::instance().read(testName);
	shared_ptr<NPtGainTestResult> result = dynamic_pointer_cast<NPtGainTestResult>(ob);
	if (!result) 
	    throw IllegalStateError("Failed to read test result: " + testName, __FILE__, __LINE__);
	if (compare(result)) {
	    if (debug) cout << "Success!" << endl;
	    return 0;
	}
	else {
	    cout << "Failed" << endl;
	    return 1;
	}
	
    } catch (Throwable& e) {
	e.sendToMrs(MRS_FATAL);
	return 1;
    }
}
